import numpy as np
import meteva
import math
from scipy.ndimage import convolve

def interpolation_linear(grd, grid, other_info='left'):
    '''
    格点到格点插值
    :param grd:左边的网格数据信息
    :param grid :右边的网格数据信息
    :other_info:网格数据除了xy方向的数值之外，还有time,dtime，leve member 等维度的值，如果other_info= 'left’则返回结果中这些维度的值就采用grd里的值，
    否则采用grid里的值，默认为：left
    :return:双线性插值之后的结果
    '''
    if (grd is None):
        return None
    # 六维转换为二维的值
    dat0 = grd.values
    dat = np.squeeze(dat0)
    grd0 = meteva.base.basicdata.get_grid_of_data(grd)
    if (grd0.dlon * grd0.nlon >= 360):
        grd1 = meteva.base.basicdata.grid([grd0.slon, grd0.elon + grd0.dlon,grd0.dlon],[grd0.slat, grd0.elat, grd0.dlat])
        dat1 = np.zeros((grd1.nlat,grd1.nlon))
        dat1[:,0:-1] = dat[:,:]
        dat1[:, -1] = dat[:, 0]
    else:
        dat1 = dat
        grd1 = grd0
    if other_info=='left':
        grd2 = meteva.base.basicdata.grid(grid.glon,grid.glat,grd0.gtime,grd0.dtimes,grd0.levels,grd0.members)
    else:
        grd2 = grid.copy()

    if (grd2.elon > grd1.elon or grd2.slon < grd1.slon or grd2.elat > grd1.elat or grd2.slat < grd1.slat):
        print("object grid is out range of original grid")
        return None

    #插值处理
    x = ((np.arange(grd2.nlon) * grd2.dlon + grd2.slon - grd1.slon) / grd1.dlon)
    ig = x[:].astype(dtype='int16')
    dx = x - ig
    y = (np.arange(grd2.nlat) * grd2.dlat + grd2.slat - grd1.slat) / grd1.dlat
    jg = y[:].astype(dtype='int16')
    dy = y[:] - jg
    ii, jj = np.meshgrid(ig, jg)
    ii1 = np.minimum(ii + 1, grd1.nlon - 1)
    jj1 = np.minimum(jj + 1, grd1.nlat - 1)
    ddx, ddy = np.meshgrid(dx, dy)
    c00 = (1 - ddx) * (1 - ddy)
    c01 = ddx * (1 - ddy)
    c10 = (1 - ddx) * ddy
    c11 = ddx * ddy
    dat2 = (c00 *dat1[jj, ii] + c10 * dat1[jj1, ii] + c01 * dat1[jj, ii1] + c11 * dat1[jj1, ii1])
    grd_new = meteva.base.basicdata.grid_data(grd2,dat2)
    return grd_new

def add(grd1,grd2,other_info = 'left'):
    '''
    将插值之后的多个结果在原有存储数据结构的基础上进行追加。
    :param grd1:左边的网格数据信息
    :param grd1 :右边的网格数据信息
    :other_info:网格数据除了xy方向的数值之外，还有time,dtime，leve member 等维度的值，如果other_info= 'left’则返回结果中这些维度的值就采用grd里的值，
    否则采用grid里的值，默认为：left
    :return:多个网格数据双线性插值之后的结果的追加
    '''
    if(grd1 is None):
        return grd2
    elif(grd2 is None):
        return grd1
    else:
        grid1 = meteva.base.basicdata.get_grid_of_data(grd1)
        grid2 = meteva.base.basicdata.get_grid_of_data(grd2)
        slon = max(grid1.slon,grid2.slon)
        elon = min(grid1.elon,grid2.elon)
        slat = max(grid1.slat,grid2.slat)
        elat = min(grid1.elat,grid2.elat)
        if other_info == 'left':
            grid_in = meteva.base.basicdata.grid([slon,elon,grid1.dlon],[slat,elat,grid1.dlat],grid1.gtime,grid1.dtimes,grid1.levels,grid1.members)
        else:
            grid_in = meteva.base.basicdata.grid([slon, elon, grid2.dlon], [slat, elat, grid2.dlat], grid2.gtime, grid2.dtimes, grid2.levels,
                              grid2.members)
        grd1_in = interpolation_linear(grd1, grid_in)
        grd2_in = interpolation_linear(grd2, grid_in)
        grd = meteva.base.basicdata.grid_data(grid_in,grd1_in.values + grd2_in.values)
        return grd


def mean_convolve(grd, half_window_size, skip=1):
    # 该函数计算网格点附近矩形方框内的平均值
    # 使用同规格的场，确保网格范围和分辨率一致
    # window_size 窗口尺度，为了避免窗口较大时计算太慢，可选择跳点取平均，再插值回到原始分辨率
    if (skip > half_window_size):
        print("pdf_skip is larger than half pdf_window_size")
        return None
    grid0 = meteva.base.get_grid_of_data(grd)
    step_num_lon = int(math.ceil((grid0.nlon - 1) / skip)) + 1
    dlon_skip = grid0.dlon * skip
    elon_skip = grid0.slon + dlon_skip * (step_num_lon - 1)
    step_num_lat = int(math.ceil((grid0.nlat - 1) / skip)) + 1
    dlat_skip = grid0.dlat * skip
    elat_skip = grid0.slat + dlat_skip * (step_num_lat - 1)
    grid_skip = meteva.base.grid([grid0.slon, elon_skip, dlon_skip],
                                                  [grid0.slat, elat_skip, dlat_skip])
    dat0 = grd.values.squeeze()
    dat = np.zeros((step_num_lat, step_num_lon))

    cycle0 = step_num_lat * step_num_lon
    cycle1 = (half_window_size * 2 + 1) * (half_window_size * 2 + 1)

    if cycle0 < cycle1:
        for j in range(step_num_lat):
            j_start = max(0,j * skip - half_window_size)
            j_end = min(grid0.nlat,j * skip + half_window_size)+1
            for i in range(step_num_lon):
                i_start = max(0, i * skip - half_window_size)
                i_end = min(grid0.nlon, i * skip + half_window_size)+1
                dat[j,i] = np.mean(dat0[j_start:j_end,i_start:i_end])
    else:
        for p in range(-half_window_size, half_window_size + 1):
            j = np.arange(step_num_lat) * skip + p
            j[j < 0] = 0
            j[j > grid0.nlat - 1] = grid0.nlat - 1
            for q in range(-half_window_size, half_window_size + 1):
                i = np.arange(step_num_lon) * skip + q
                i[i < 0] = 0
                i[i > grid0.nlon - 1] = grid0.nlon - 1
                ii, jj = np.meshgrid(i, j)
                dat += dat0[jj, ii]
        dat /= cycle1

    grd_mean_skip = meteva.base.grid_data(grid_skip, dat)
    grd_mean = meteva.base.function.gxy_gxy.interpolation_linear(grd_mean_skip, grid0)

    return grd_mean


def smooth(grd,times = 1):
    dat = grd.values.squeeze()
    kernel = np.array([[0.0625, 0.125, 0.0625],
                       [0.125, 0.25, 0.125],
                       [0.0625, 0.125, 0.0625]])
    for i in range(times):
        dat = convolve(dat, kernel)

    grid0 = meteva.base.get_grid_of_data(grd)
    grd1 = meteva.base.grid_data(grid0,dat)
    return grd1


# -*- coding:utf-8 -*-
import os
from concurrent import futures
from multiprocessing import Pool

def mult_process(func=None, func_args_all=[], max_workers=4, force_max_workers=False):
    '''
    [多进程计算/绘图]
    Keyword Arguments:
        func {[type]} -- [函数] (default: {None})
        func_args_all {list} -- [函数参数] (default: {[]})
        max_workers {number} -- [进程池大小] (default: {6})
        force_max_workers {bool} -- [是否强制使用max_workers参数，默认关闭] (default: {False})
    Returns:
        [list] -- [有效的返回结果]
    '''
    if len(func_args_all) == 0:
        return []

    # 多进程绘图
    if force_max_workers == False:
        use_sys_cpu = max(1, os.cpu_count() - 1)  # 可使用的cpu核心数为cpu核心数-1
        max_workers = min(max_workers, use_sys_cpu)  # 最大进程数
    print('cpu_count={}, use max_workers={}'.format(os.cpu_count(), max_workers))

    all_ret = []
    with futures.ProcessPoolExecutor(max_workers=max_workers) as executer:
        # 提交所有绘图任务
        all_task = [executer.submit(func, _) for _ in func_args_all]
        # 等待
        futures.wait(all_task, return_when=futures.ALL_COMPLETED)
        # 取返回值
        for task in all_task:
            exp = task.exception()
            if exp is None:
                all_ret.append(task.result())
            else:
                print(exp)
                pass
    return all_ret    


## 并行计算相关
def __split_list_equal(list0, n):
    ## list0等分，每份n个
    for i in range(0, len(list0),n):
        yield list0[i:i+n]

def __split_list_nlist(list0, n):
    ## list0等分为n组
    if len(list0)%n == 0:
        cnt = len(list0)//n
    else:
        cnt = len(list0)//n+1
    for i in range(0,n):
        yield list0[i*cnt : (i+1)*cnt]

def multi_pool_cal(operation, input, pro_count):
    """
    不带返回值的并行同步处理
    ## operation为待并行函数
    ## input为某参数并行列表(list)， pro_count为进程数
    ## 根据pro_count自动将input切分为等长的n份，作为并行参数
    """
    processes_pool = Pool(pro_count)
    input_mpi = list(__split_list_nlist(input, pro_count))
    # 开始并行
    processes_pool.map(operation, input_mpi)
    return None


def multi_model_process(model, input, n_pools = 6):
    """
    带返回值的并行同步处理，返回多结果列表。
    ## operation为待并行函数，支持多参数(args)
    ## 进行异步非阻塞循环
    """
    results=[]
    datelist_mpi = list(__split_list_nlist(input, n_pools))#大list平分为若干小list
    processes_pool = Pool(n_pools)
    for dtlist in datelist_mpi:
        print(dtlist)
        res = processes_pool.apply_async(model.process, args=(dtlist,))
        results.append(res)
    processes_pool.close()
    processes_pool.join()
    a = []
    # 数据整理,apply_async()组织数据
    a = [i.get() for i in results]
    results = pd.concat(a, axis=0)
    return(results)


if __name__ =="__main__":
    import numpy as np
    #result = task(numbers)
    list0 = np.arange(10)
    print(list(__split_list_equal(list0, 3)))
    pass


import numpy as np
import meteva
import pandas as  pd
def mean_of_sta_ensemble(sta_ensemble):
    sta_mean = sta_ensemble[meteva.base.get_coord_names()]
    sta_data = sta_ensemble[meteva.base.get_data_names(sta_ensemble)]
    value = sta_data.values
    mean = np.mean(value,axis=1)
    sta_mean['data0'] =mean
    return sta_mean

def std_of_sta_ensemble(sta_ensemble):
    sta_std = sta_ensemble[meteva.base.get_coord_names()]
    sta_data = sta_ensemble[meteva.base.get_data_names(sta_ensemble)]
    value = sta_data.values
    std = np.std(value, axis=1)
    sta_std['data0'] = std
    return sta_std

def var_of_sta_ensemble(sta_ensemble):
    sta_var = sta_ensemble[meteva.base.get_coord_names()]
    sta_data = sta_ensemble[meteva.base.get_data_names(sta_ensemble)]
    value = sta_data.values
    var = np.var(value, axis=1)
    sta_var['data0'] = var
    return sta_var


def max_of_sta_ensemble(sta_ensemble):
    sta_max = sta_ensemble[meteva.base.get_coord_names()]
    sta_data = sta_ensemble[meteva.base.get_data_names(sta_ensemble)]
    value = sta_data.values
    max1 = np.max(value, axis=1)
    sta_max['data0'] = max1
    return sta_max

def min_fo_sta_ensemble(sta_ensemble):
    sta_min = sta_ensemble[meteva.base.get_coord_names()]
    sta_data = sta_ensemble[meteva.base.get_data_names(sta_ensemble)]
    value = sta_data.values
    min1 = np.min(value, axis=1)
    sta_min['data0'] = min1
    return sta_min

def get_one_element_by_name(sta_ensemble,name):
    column = meteva.base.get_coord_names()
    column.append(name)
    sta = sta_ensemble[column]
    return sta

def get_one_element_by_number(sta_ensemble,number):
    column = meteva.base.get_coord_names()
    data_names = meteva.base.get_data_names(sta_ensemble)
    column.append(data_names[number])
    sta = sta_ensemble[column]
    return sta


import os
import meteva_base as meb
import bz2
import logging
import time
import numpy as np
import pygrib
import pandas as pd
import xarray as xr


def un_bz2_file(file_name):
    """
    解压bz2 数据,生成文件
    :param file_name:文件绝对路径
    :return: 解压后文件名
    """
    if file_name.endswith('.bz2'):
        if not os.path.exists(file_name.split(".")[0]):
            st = time.time()
            newfilepath = os.path.join(file_name.split(".")[0])
            try:
                with open(newfilepath, 'wb+') as new_file, bz2.BZ2File(file_name, 'rb') as file:
                    for data in iter(lambda: file.read(10000 * 1024), b''):
                        new_file.write(data)
                print('decompress:' + newfilepath + ' with time:' + str(time.time() - st) + 's')
                return file_name.split(".")[0]
            except Exception as e:
                print(file_name + " data exception, please check")
                return None

def un_bz2_handle(file_name):
    """
    解压bz2 数据,生成文件obj
    :param file_name:文件绝对路径
    :return: 解压后文件名
    """
    if file_name.endswith('.bz2'):
        if not os.path.exists(file_name.split(".")[0]):
            st = time.time()
            try:
                file = bz2.BZ2File(file_name, 'rb')
                return(file)
            except Exception as e:
                print(file_name + " data exception, please check")
                return None

def dateDiffInHours(t1, t2):
    """
    计算预报时效
    :param t1:起报时间
    :param t2:预报时间
    :return: 小时数
    """
    td = t2 - t1
    return int(td.days * 24 + td.seconds / 3600)


def _is_nonstring_container(obj):
    """
    判断是否是非字符串的 container容器类。 返回bool类型
    """
    from collections.abc import Container, Iterable
    return isinstance(obj, Container) and isinstance(obj, Iterable) and not isinstance(obj, (str, bytes, bytearray))


def _pygrib_param_id(grb):
    """
    返回grib_message的param_id
    """
    try:
        # 对grib1类型数据获取其表格编码，两组数字组成
        param_id = str.format("{0:0>3d}_{1:0>3d}", grb.table2Version, grb.paramId)
    except Exception as e:
        try:
            # 对grib2类型数据提取，三组数字组成
            param_id = str.format("{0}_{1}_{2}", grb.discipline, grb.parameterCategory, grb.parameterNumber)
        except Exception as err:
            msg = r"ERROR: no grib param_id, please CHECK! "
            print(msg)
            raise ValueError(err)
    return param_id

def pygrib_vars_info(f, save_path=None):
    """
    读取grib数据的变量清单,pygrib库依赖
    :param 
        f: pygrib.open()打开文件对象
        save_path: 变量清单结果的存储路径,csv格式；默认为None不保存
    :return: 变量清单(pandas.DataFrame)
    """
    f.seek(0,0)
    data_list = []
    for grb in f:
        # try:
        #     # 对grib1类型数据获取其表格编码，两组数字组成
        #     param_id = str.format("{0:0>3d}_{1:0>3d}", grb.table2Version, grb.paramId)
        # except Exception as e:
        #     # 对grib2类型数据提取，三组数字组成
        #     param_id = str.format("{0}_{1}_{2}", grb.discipline, grb.parameterCategory, grb.parameterNumber)
        param_id = _pygrib_param_id(grb)
    #     print(grb.shortName, grb.name, param_id, grb.typeOfLevel, grb.level)
        data = pd.DataFrame([[grb.shortName, grb.name, param_id, grb.typeOfLevel, grb.level]],
                            columns=['shortName', 'name', 'param_id', 'typeOfLevel', 'level' ])
        data_list.append(data)
    data_all = pd.concat(data_list, axis=0)
    ## 数据去重排序
    data_all.drop_duplicates(keep='first', inplace=True)
    data_all.sort_values(by=['typeOfLevel','shortName','level'],inplace=True)
    if save_path is not None: ##文件保存变量清单
        try:
            data_all.to_csv(save_path)
        except Exception as err:
            print(err)
    return(data_all)

def read_pygrib_message(f, shortName=None, typeOfLevel=None, level=None, id=None):
    """
    读取grib变量并输出,pygrib库依赖
    :param 
        f: pygrib.open()打开文件对象
        shortName: 要素变量的简写名，具体可查看变量清单。支持str及多str列表形式，用于提取一个或多个要素；默认为None提取所有要素
        typeOfLevel: 要素为单层或高空多层，具体可查看变量清单。('isobaricInhPa','surface',...)
        level: 要素层次，typeOfLevel为'isobaricInhPa'时可设置。支持int及int列表形式；默认为None提取所有层次
        id   : param_id（grib/grib2），具体见_pygrib_param_id（）函数
    :return: 变量(xarray.DataArray)列表list
    """
    dict0 = {}## select 参数列表
    if shortName is not None:
        dict0['shortName'] = shortName
    if typeOfLevel is not None:
        dict0['typeOfLevel'] = typeOfLevel
    if level is not None:
        if typeOfLevel != 'isobaricInhPa':
            print("Warning: CHECK whether typeOfLevel == isobaricInhPa if LEVEL is not None")
        dict0['level'] = level

    
    # 解码并存放读取完成的要素
    array = []
    s = time.time()
    ##pygrib.open/index为相似功能类；其中index.select参数只限制于单要素，open.select参数可以包含容器。 速度index.select>>open.select
    try:
        grib_data = f.select(**dict0)
    except ValueError as e:
        logging.info(e)
        raise ValueError(e)
    ## 循环解码要素列表
    for i in grib_data:
        try:
            # 混合编码数据可能会抛出的异常
            # try:
            #     # 对grib1类型数据获取其表格编码，两组数字组成
            #     param_id = str.format("{0:0>3d}_{1:0>3d}", i.table2Version, i.paramId)
            # except Exception as e:
            #     # 对grib2类型数据提取，三组数字组成
            #     param_id = str.format("{0}_{1}_{2}", i.discipline, i.parameterCategory, i.parameterNumber)
            param_id = _pygrib_param_id(i)
            if id is not None:
                if _is_nonstring_container(id): # param_id改为数组
                    id = [id]
                if param_id not in id:# 有id参数时，只解码对应param_id是否在id列表中
                    continue
            print('param_id', param_id)
            valid_time = meb.all_type_time_to_datetime(str(i.validityDate) + "{0:0>4d}".format(i.validityTime))
            data_time = meb.all_type_time_to_datetime(str(i.dataDate) + "{0:0>4d}".format(i.dataTime))
            # data_time = i.validDate
            # fhour = i.forecastTime
            fhour0 = dateDiffInHours(data_time, valid_time)
            # print(valid_time, data_time, fhour0)
            level = i.level
            s_lat = i.latitudeOfFirstGridPointInDegrees  # 纬度起始值
            e_lat = i.latitudeOfLastGridPointInDegrees  # 纬度结束值
            step_lat = i.jDirectionIncrementInDegrees  # 纬度间隔
            s_lon = i.longitudeOfFirstGridPointInDegrees  # 经度起始值
            e_lon = i.longitudeOfLastGridPointInDegrees  # 经度结束值
            if (s_lon>e_lon) and s_lon==180.:
                s_lon = 0-s_lon 
            step_lon = i.iDirectionIncrementInDegrees  # 经度步长
            latlon = i.latlons()
            lats = latlon[0][:,0]
            lons = latlon[1][0,:]
            # print(lons, lats)
            # print(latlon[0], latlon[1])
            # print( latlon[0].shape, latlon[1].shape)
            g_lon = [s_lon, e_lon, -step_lon if s_lon > e_lon else step_lon]
            g_lat = [s_lat, e_lat, -step_lat if s_lat > e_lat else step_lat]
            g_time = [data_time]
            # print(s_lon, e_lon, step_lon, s_lat, e_lat, step_lat)
            d_time = [dateDiffInHours(data_time, valid_time)]
            member = param_id + '-' + i.shortName + '-' + str(i.level)
            grid_info = meb.grid(glon=g_lon, glat=g_lat, gtime=g_time,
                                dtime_list=d_time, level_list=[level],
                                member_list=[member])
            grd = meb.grid_data(grid_info, i.values)
            grd.name = i.shortName
            # print(param_id, g_lat, g_lon, level, valid_time, data_time, d_time)
            array.append(grd)
        except Exception as e:
            logging.info(e)
            print('ERROR', e)
            continue
    # keys = i.keys()
    # lonlat = i.latlons()
    # print(i.validityDate, i.validityTime, i.validDate,  i.significanceOfReferenceTime, i.dataDate ,i.dataTime, i.forecastTime, i.julianDay)
    # print(i.year, i.month, i.day, i.hour, i.minute, i.second)
    f.close()
    logging.info('Data read complete with time:%s',  str(time.time() - s))
    return array


def read_pygrib(file_path,  shortName=None, typeOfLevel=None, level=None, id=None,
            show_vars=False, dset=False):
    """
    读取grib格式数据，pygrib依赖
    :param 
        file_path: 数据文件绝对路径
        shortName: 要素变量的简写名，具体可查看变量清单。支持str及多str列表形式，用于提取一个或多个要素；默认为None提取所有要素
        typeOfLevel: 要素为单层或高空多层，具体可查看变量清单。('isobaricInhPa','surface',...)
        level: 要素层次，typeOfLevel为'isobaricInhPa'时可设置。支持int及int列表形式；默认为None提取所有层次
        show_vars: 是否返回要素清单，默认为否。可用该参数先了解grib文件中要素名称、种类等，再进行具体提取
        dset: 是否返回xarray.DataSet数据，默认为否； False:返回DataArray列表list
    :return: 返回提取要素，网格数据(xarray.DataSet或xarray.DataArray列表)
    """
    np.set_printoptions(suppress=True)

    # 解压bz2文件
    if not os.path.exists(file_path):
        print('File missing {0}'.format(file_path))
        return None
    try:
        if file_path.endswith('.bz2'):#bz2文件
            file_path0 = os.path.join(os.path.dirname(file_path), os.path.splitext(file_path)[0])#解压文件名
            print(file_path0)
            if os.path.exists(file_path0):#存在解压文件
                file_name = file_path0
            else:
                print("OPEN {}: Compressed".format(file_name))
                file_name = un_bz2_file(file_path)
        else:#非压缩文件名
            print("OPEN {}: not Compressed".format(file_path))
            file_name = file_path
    except Exception as err:
        print(err)
        return None
    # 读取文件
    try:
        pygrib.multi_support_on()
        f = pygrib.open(filename=file_name)
    except Exception as err:
        print(err)
        return None

    # 显示变量vars清单
    if show_vars == True:
        df_vars = pygrib_vars_info(f)
        return df_vars
    else:
        # 解码并存放读取完成的要素
        array = read_pygrib_message(f, shortName=shortName, typeOfLevel=typeOfLevel, level=level, id=id)
        if dset==False:
            return array
        else:
            arrays = xr.merge(array)
            return arrays


def get_stainfo(lons, lats, ids=None):
    if ids is None:
        ids = np.arange(len(lons))
    data = {"站号": ids,
            "经度": lons,
            "纬度": lats,}
    df = pd.DataFrame(data)
    sta = meb.sta_data(df,columns = ["id","lon","lat"])
    print(sta)
    return sta

def read_sta_from_grib2(file, sta_info):
    array = read_pygrib(file)
    grd = xr.merge(array)['data0']
    result = meb.interp_gs_linear(grd, sta=sta_info)
    return result, grd

if __name__ == '__main__':
    ## 02 SCMOC_ER01
    lons = [116.35]
    lats = [39.95]
    sta_info = get_stainfo(lons=lons, lats=lats)
    file = r"J:\Z_NWGD_C_BABJ_20230505055246_P_RFFC_SCMOC-ER01_202305050500_03601.GRB2"
    array = read_pygrib(file, dset=True)#返回xr.dataset数据
    print(array)
    sta = meb.interp_gs_linear(array.unknown, sta=sta_info)
    print(sta)



#!/usr/bin/python3.6
# -*- coding:UTF-8 -*-
import datetime as datetime
import os as os
import numpy as np
# from ..io import DataBlock_pb2
# from ..io.GDS_data_service import GDSDataService
# import meteva
import meteva_base
import re
import pathlib

#获取最新的路径


#获取最新的gds路径
def get_latest_gds_path(dir,time,dtime,dt_cell = "hour",dt_step = 1,farthest = 240):
    dir1,filemodel = os.path.split(dir)
    file_list = get_gds_file_list_in_one_dir(dir1)
    for ddt in range(0,farthest,dt_step):
        if dt_cell.lower() == "hour":
            time1 = time - datetime.timedelta(hours=ddt)
        elif dt_cell.lower() == "minute":
            time1 = time - datetime.timedelta(minutes=ddt)
        filename = get_path(filemodel, time1, dtime + ddt, dt_cell)
        if filename in file_list:
            path = dir1 + "/" + filename
            return path

#获取最新的路径
def get_latest_path(dir,time,dt,dt_cell = "hour",dt_step = 1, farthest = 240):
    for ddt in range(0,240,dt_step):
        if dt_cell.lower()=="hour":
            time1 = time - datetime.timedelta(hours=ddt)
        elif dt_cell.lower()=="minute":
            time1 = time - datetime.timedelta(minutes=ddt)

        path = get_path(dir,time1,dt + ddt,dt_cell)
        if(os.path.exists(path)):
            return path
    return None


#获取路径
def get_path(dir,time,dt = None,dt_cell = "hour"):
    '''
    根据路径通配形式返回路径
    :param dir:
    :param time:
    :param dt:
    :param dt_cell:
    :return:
    '''
    if(dir.find("*")<0):
        return get_path_without_star(dir,time,dt,dt_cell)
    else:
        dir_0,filename_0 = os.path.split(dir)
        dir_1 = get_path_without_star(dir_0,time,dt,dt_cell)
        filename_1 = get_path_without_star(filename_0,time,dt,dt_cell)
        patten = dir_1  +"\\"+ filename_1
        patten = patten.replace("\\","/")
        patten = patten.replace("*","\\S+")
        if os.path.exists(dir_1):
            files = os.listdir(dir_1)
            for file in files:
                path = dir_1 + "/" + file
                path = path.replace("\\","/")
                match = re.match(patten,path)
                if match is not None:
                    return path
    return None



def get_path_without_star(dir,time,dt = None,dt_cell = "hour"):
    '''
    :param dir:
    :param time:
    :param dt:
    :param dt_cell:
    :return:
    '''
    if(dt is not None):
        if not (isinstance(dt,np.int16) or isinstance(dt,np.int32) or isinstance(dt,np.int64)or type(dt) == type(1)):
            if(dt_cell.lower()=="hour"):
                dt = int(dt.total_seconds() / 3600)
            elif(dt_cell.lower()=="minute"):
                dt = int(dt.total_seconds() / 60)
            elif(dt_cell.lower()=="day"):
                dt = int(dt.total_seconds() / (24*3600))
            else:
                dt = int(dt.total_seconds())
    else:
        dt = 0
    cdt2 = '%02d' % dt
    cdt3 = '%03d' % dt
    cdt4 = '%04d' % dt
    dir1 = dir.replace("TTTT",cdt4).replace("TTT",cdt3).replace("TT",cdt2)
    y4 = time.strftime("%Y")
    y2 = y4[2:]
    mo = time.strftime("%m")
    dd = time.strftime("%d")
    hh = time.strftime("%H")
    mi = time.strftime("%M")
    ss = time.strftime("%S")
    dir1 = dir1.replace("YYYY",y4).replace("YY",y2).replace("MM",mo).replace("DD",dd).replace("HH",hh).replace("FF",mi).replace("SS",ss)
    dir1 = dir1.replace(">","")
    return dir1

#创建路径
def creat_path(path):
    if path is not None:
        [dir,filename] = os.path.split(path)
        #if(not os.path.exists(dir)):
        #    os.makedirs(dir)
        pathlib.Path(dir).mkdir(parents=True,exist_ok=True)


#字符转换为datetime
def str_to_time(str0):
    return datetime.datetime.strptime(str0, '%Y%m%d%H%M')

#获取预报时效路径
def get_forecat_hour_of_path(path_model,path):
    ttt_index = path_model.find("TTT")
    if (ttt_index >= 0):
        ttt = int(path[ttt_index:ttt_index + 3])
    else:
        ttt = 0
    return ttt

#获取时间命名的路径
def get_time_of_path(path_model,path):

    yy_index = path_model.find("YYYY")
    if  yy_index < 0:
        yy_index = path_model.find("YY")
        if(yy_index <0):
            yy = 2000
        else:
            yy = int(path[yy_index: yy_index + 2])
    else:
        yy = int(path[yy_index: yy_index+4])

    mm_index = path_model.find("MM")
    if(mm_index >=0):
        mm = int(path[mm_index:mm_index+2])
    else:
        mm = 1

    dd_index = path_model.find("DD")
    if(dd_index>=0):
        dd = int(path[dd_index:dd_index + 2])
    else:
        dd = 1
    hh_index = path_model.find("HH")
    if(hh_index>=0):
        hh = int(path[hh_index:hh_index + 2])
    else:
        hh = 0
    ff_index = path_model.find("FF")
    if(ff_index>=0):
        ff = int(path[ff_index:ff_index + 2])
    else:
        ff = 0
    ss_index = path_model.find("SS")
    if(ss_index>=0):
        ss = int(path[ss_index:ss_index + 2])
    else:
        ss = 0
    return datetime.datetime(yy,mm,dd,hh,ff,ss)

#获取按照时间命名的文件夹名称
def get_dir_of_time(path_model,time):
    dir = os.path.split(get_path(path_model,time))[0]
    return dir

#获取按照时间命名的文件夹名称列表
def get_path_list_of_time(path_model,time):
    dir = get_dir_of_time(path_model,time) +"/"
    path_list = []
    if(os.path.exists(dir)):
        path_list = os.listdir(dir)
        for i in range(len(path_list)):
            path_list[i] = dir  + path_list[i]
    return path_list

#按照最新的时间命名文件路径
def get_time_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = abs((time1 - time).total_seconds())
        if(dt < dt_min):
            dt_min = dt
            nearest_path = path
    return nearest_path

#按照上一个时效命名文件路径
def get_time_before_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = (time - time1).total_seconds()
        if(dt < dt_min and dt >=0):
            dt_min = dt
            nearest_path = path
    return nearest_path

##按照下一个时效命名文件路径
def get_time_after_nearest_path(path_model,time,max_seconds,path_list = None):
    if(path_list is None):
        path_list = get_path_list_of_time(path_model,time)
    dt_min = max_seconds
    nearest_path = None
    for path in path_list:
        time1 = get_time_of_path(path_model,path)
        dt = (time1 - time).total_seconds()
        if(dt < dt_min  and dt >=0):
            dt_min = dt
            nearest_path = path
    return nearest_path

#以列表的形式返回根目录下所有文件的路径
def get_path_list_in_dir(root_dir,all_path = None):
    if not os.path.exists(root_dir):
        return []
    files = os.listdir(root_dir)
    if all_path is None:
        all_path = []
    for file in files:
        fi_d = os.path.join(root_dir,file)
        if os.path.isdir(fi_d):
            get_path_list_in_dir(fi_d,all_path)
        else:
            all_path.append(fi_d)

    return all_path

#返回根目录下nc格点文件的长名字
def get_path_of_grd_nc_longname(root_dir,time,dhour,nc_Fname,fhour_add):
    ruc_file = "{4}_IT_{0}_VT_{1}_FH_{2:0>3d}_AT_{3:0>3d}.nc".format(time.strftime("%Y%m%d%H"),
                                                                     (time + datetime.timedelta(hours=int(dhour))).strftime(
                                                                         "%Y%m%d%H"), dhour, fhour_add, nc_Fname)
    file = r"{0}\{1}\{2}".format(root_dir, time.strftime("%Y%m%d"), ruc_file)
    return file


def get_gds_file_list_in_one_dir(dir):
    dir = dir.replace("mdfs:///", "")

    ip,port = meteva_base.gds_ip_port
    service = GDSDataService(ip, port)
    # 获得指定目录下的所有文件
    status, response = service.getFileList(dir)
    MappingResult = DataBlock_pb2.MapResult()
    file_list = []
    if status == 200:
        if MappingResult is not None:
            # Protobuf的解析
            MappingResult.ParseFromString(response)
            results = MappingResult.resultMap
            # 遍历指定目录
            for name_size_pair in results.items():
                if (name_size_pair[1] != 'D'):
                    file_list.append(name_size_pair[0])
    return file_list

def get_gds_path_list_in_one_dir(dir):
    file_list = get_gds_file_list_in_one_dir(dir)
    path_list = []
    for file in file_list:
        path_list.append(dir + "/"+ file)
    return path_list

def exist_in_gds(path):
    dir,filename = os.path.split(path)
    file_list = get_gds_file_list_in_one_dir(dir)
    if filename in file_list:
        return True
    else:
        return False

def get_gds_all_dir(path,all_path,service = None):
    # 初始化GDS客户端

    if service is None:
        ip, port = meteva_base.gds_ip_port
        service = GDSDataService(ip, port)
    # 获得指定目录下的所有文件
    path = path.replace("mdfs:///", "")
    status, response = service.getFileList(path)
    MappingResult = DataBlock_pb2.MapResult()
    if status == 200:
        if MappingResult is not None:
            # Protobuf的解析
            MappingResult.ParseFromString(response)
            results = MappingResult.resultMap
            # 遍历指定目录
            contain_dir = False
            for name_size_pair in results.items():
                if (name_size_pair[1] == 'D'):
                    contain_dir = True
                    path1 = '%s%s%s' % (path, "/" , name_size_pair[0])
                    if(path1[0:1] == "/"):
                        path1 = path1[1:]
                    if(path1[0:1] == "/"):
                        path1 = path1[1:]
                    get_gds_all_dir(path1,all_path,service)
                    #print(name_size_pair[0])
            if(not contain_dir):
                all_path.append(path)



def get_dati_of_path(path):
    try:
        dir,filename = os.path.split(path)
        filename0 = os.path.splitext(filename)[0]
        a = int(filename0[0:2])
        b = int(filename0[2:4])
        if a ==20:
            if b >12:
                pass
            else:
                filename0 = "20" + filename0
        elif a == 19:
            if b >12:
                pass
            else:
                filename0 = "20" + filename0
        else:
            filename0 = "20" + filename0

        dati = str_to_time(filename0)
        return dati
    except:
        return None

#以列表的形式返回根目录下所有文件的路径
def get_during_path_list_in_dir(root_dir,all_path = None,start = None,end = None):
    start_time = None
    if start is None:
        start_time = datetime.datetime(1900,1,1,0,0)
    else:
        start_time = start
    if end is None:
        end_time = datetime.datetime.now()
    else:
        end_time = end
    time_compair = False
    if start is not None or end is not None:
        time_compair = True

    if not os.path.exists(root_dir):
        return []
    files = os.listdir(root_dir)
    if all_path is None:
        all_path = []
    for file in files:
        fi_d = os.path.join(root_dir,file)
        if os.path.isdir(fi_d):
            if time_compair:
                dati = get_dati_of_path(fi_d)
                if dati is None:
                    get_during_path_list_in_dir(fi_d, all_path, start, end)
                else:
                    if dati >= start_time and dati <= end_time:
                        get_during_path_list_in_dir(fi_d,all_path,start,end)
            else:
                get_during_path_list_in_dir(fi_d, all_path, start, end)
        else:
            all_path.append(fi_d)
    return all_path

# 返回文件名中对应的预报时间时效等信息
def get_time_from_path(fmt=r"\\10.28.16.179\rucdata2\SC\HRCLDAS\T2\YYYY\YYYYMMDD\YYYYMMDDHHFF.nc" , 
                    filename=r"\\10.28.16.179\rucdata2\SC\HRCLDAS\T2\2023\20230516\202305161000.nc"):
    mo = 1
    dd = 1
    hh = 0
    mm = 0
    ss = 0
    try:
        ## TTT
        if fmt.find('TTTT') >=0:
            cdt = int(filename[fmt.find('TTTT'):fmt.find('TTTT')+4])
        elif fmt.find('TTT') >=0:
            cdt = int(filename[fmt.find('TTT'):fmt.find('TTT')+3])
        elif fmt.find('TT') >=0:
            cdt = int(filename[fmt.find('TT'):fmt.find('TT')+2])
        else:
            cdt = 24
        ## year
        if fmt.find('YYYY') >=0:
            year = int(filename[fmt.find('YYYY'):fmt.find('YYYY')+4])
        elif fmt.find('YY') >=0:
            year = 2000 + int(filename[fmt.find('YY'):fmt.find('YY')+2])
        else:
            year = 2020
        ## month
        mo = int(filename[fmt.find('MM'):fmt.find('MM')+2])
        ## day
        dd = int(filename[fmt.find('DD'):fmt.find('DD')+2])
        ## time
        hh = int(filename[fmt.find('HH'):fmt.find('HH')+2])
        mm = int(filename[fmt.find('FF'):fmt.find('FF')+2])
        ss = int(filename[fmt.find('SS'):fmt.find('SS')+2])
    except Exception as err:
        print(err)
    time  = datetime(year=year, month=mo, day=dd, hour=hh, minute=mm, second=ss)
    dtime = cdt
    return time, dtime


######### 文件操作 ######## 
def _is_file_copy_by_time(source_file,local_file):
    ## 判断源文件和转换文件关系，若源文件存在，转换文件不存在或修改时间早于源文件，则返回True   
    if os.path.exists(source_file):
        if not os.path.exists(local_file) or (os.path.getmtime(local_file)<os.path.getmtime(source_file)):
            return(True)#满足上述条件，可以转换，返回TURE
    return(False)


def copy_file(source_file,local_file):
    ## 复制本地文件至目标文件，如果目标文件不存在或修改时间早于本地文件，则启动文件夹创建+复制
    import shutil
    if _is_file_copy_by_time(source_file,local_file):
        #确认需复制
        #文件夹
        path = os.path.split(local_file)[0]
        if not os.path.exists(path) : os.makedirs(path)
        #文件复制
        shutil.copyfile(source_file,local_file)
        print("COPY: {0} - {1}".format(source_file,local_file))
    return()


############# 文件BackUP相关
def _check_all_files(fmt,time,dtimes):
    # 检验时间时效对应的通配文件，是否全部存在
    # 当dtime<0 时， 不检查直接通过
    # 返回为True即该时效列表对应文件全存在，False则为不全存在
    flag = True 
    miss = []
    for dtime in dtimes:  
        if dtime < 0 :
            continue      
        filename = meb.get_path(fmt, time=time, dt=dtime)
        if not os.path.exists(filename):
            # print("FILE not Existed: {}".format(filename))
            flag = False
            miss.append(dtime)
    miss = np.array(miss)
    return(flag, miss)

## 备份中与起报时间对应的预报时效的转换
def _dt_fh_cal(dt_raw,dt_new,fh_raw,unit='hour', keep_max=False):
    ## 变换生成备份后fh列表,
    ## keep_max为True时，保证最大、最小值与原列表一致
    if unit == 'hour':
        delta_hr = int((dt_new-dt_raw).total_seconds()/3600)
    elif unit == 'minute':
        delta_hr = int((dt_new-dt_raw).total_seconds()/60)
    if type(fh_raw) is not np.ndarray:
        print(type(fh_raw))
        fh_raw = np.array(fh_raw)
    fh_new = fh_raw-delta_hr
    # print(fh_new)
    ## 小于0时效赋值为NaN
    fh_new[fh_new<0] = -1
    if keep_max:
        fh_new[fh_new<0] = np.min(fh_new[fh_new>=0])
    return fh_new

## 备份文件
def _copyfiles_time_dtimes(time0, dtimes0, fmt0,
                            time1, dtimes1, fmt1,
                            func = None, show=False, **args):
    import shutil
    ## 0为目的文件(copy to)， 1为源文件(待copy)
    ## dtime1待备份文件的预报时效<0时， 直接跳过不复制
    ## func: 复制外的其他处理，可用func函数定义: func(file1, file0)#读取file1， 输出file0
    if not len(dtimes0) == len(dtimes1):
        raise IOError("copy dtimes_length must the SAME")
        return None
    for dtime0, dtime1 in zip(dtimes0, dtimes1):
        if dtime1 < 0:
            continue
        else:
            file0 = meb.get_path(fmt0, time=time0, dt=dtime0)#copy to
            file1 = meb.get_path(fmt1, time=time1, dt=dtime1)#copy from
            if not os.path.exists(file1):
                print('FILE not EXISTED: {0}'.format(file1))
                continue
            if func is None:
                func = copy_file
            if show:
                print("FROM {0}, TO {1}".format(file1, file0))
            dirname = os.path.dirname(file0)
            # if not os.path.exists(dirname): os.makedirs(dirname)
            func(file1, file0, **args)
    return None


def get_bakup_path_in_time_dtime(time, dtimes, fmt, 
                    fmt_bak, bak_h=10, delta=1, unit='hour',
                    func=None, show=True, **args):
    import copy
    ## 从备份文件中，挑选满足
    ## bak process
    # bak_h: 向前寻找的time length
    # delta：向前寻找time_delta,  unit: 'hour' or 'minute'
    # args : func parameters
    time_raw = copy.copy(time)
    print('Processing Datetime: {0}'.format(time_raw))
    ## raw file existed?
    result = _check_all_files(fmt,time,dtimes)
    if result[0] is True:
        print("FILES EXIST, no Backup: {0}".format(time_raw))
        return None
    dtimes_raw = result[1]# miss file dtimes
    delta_time = delta
    while delta_time <= bak_h:
        if unit=='hour':
            time_bak = time - datetime.timedelta(hours=delta_time) #hour
        else:
            time_bak = time - datetime.timedelta(minutes=delta_time) #minute
        dtimes_bak = _dt_fh_cal(time_raw, time_bak, dtimes_raw, unit=unit)
        ## check files existed
        result = _check_all_files(fmt_bak, time_bak, dtimes_bak)
        if result[0] is False:# bak file not complete  
            if show : print("BAK TIME miss FILES: {0}. CONTINUE!".format(time_bak))
            delta_time += delta
            continue 
        else:# bak file complete
            if show : print("BAK TIME all EXIST : {0}. Start PROCESS!".format(time_bak))
            _copyfiles_time_dtimes(time_raw, dtimes_raw, fmt,
                            time_bak, dtimes_bak, fmt_bak, 
                            func=func, show=show, **args)
            print("SUCCESS BAK: {0} - {1}".format(time_bak, dtimes_bak))
            return None
    print("File not Matching: {0}".format(time_raw))
    return None


if __name__ == "__main__":
    pass


# -*- coding:UTF-8 -*-
import copy
import numpy as np
import meteva_base
import pandas as pd

def sta_data(df,columns = None,
             # dtime_units="hour", data_source="", level_type="",
             # var_name="", var_cn_name="",
             # var_units="", valid_time=0, data_start_columns=6,
             reset_ids = True,
             units_attr      = '',#数据单位
             model_var_attr  = '',#补充模式/要素信息
             dtime_units_attr= 'hour',# hour/minite
             level_type_attr = 'isobaric',# isobaric/attitude
             time_type_attr  = 'UT',#UT/BT
             time_bounds_attr= [0,0],#数据起止时间,
             is_dtype_conv   = True
             ):
    '''
    sta_data() 对数据进行格式化成为固定格式
    :param df: dataframe的站点数据
    :param columns: 文件内包含的数据的列名
    :return: 包含‘level', 'time', 'dtime', 'id', 'lon', 'lat',  列的一个dataframe
    '''
    #提取dframe0 列名称
    if columns is None:
        columns = df.columns

    # 将列名变为小写
    columns_1 = []
    for column in columns:
        #column = column.lower()
        columns_1.append(column)
    columns = columns_1

    new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']

    # 提取数据列名称,扩展到新df的列名称中
    for column in columns:
        if column not in new_columns:
            new_columns.append(column)
    sta = copy.deepcopy(df)

    sta.columns = columns
    if "id" not in columns:
        sta["id"] = meteva_base.IV
    if reset_ids :reset_id(sta)
    sta.reset_index(inplace=True)

    # 更改列名
    sta = sta.reindex(columns = new_columns)
    #dframe1 = dframe1[new_columns]

    # 排序
    sta.sort_values(by=new_columns[:4],inplace=False)

    if len(sta.columns) == 6:
        sta["data0"] =0
    else:
        #for i in range(6,len(sta.columns)):
        #    sta.iloc[:,i] = (sta.values[:,i]).astype(np.float32)
        pass

    set_stadata_attrs(sta,units = units_attr, model_var = model_var_attr, dtime_units = dtime_units_attr,
                      level_type = level_type_attr, time_type = time_type_attr, time_bounds = time_bounds_attr)
    if is_dtype_conv:
        sta=set_stadata_coords_dtype(sta)
    return sta

def set_stadata_attrs(sta, units = None, model_var = None, dtime_units =None,
             level_type = None, time_type = None,
             time_bounds = None):

    if sta.attrs is None: sta.attrs = {}
    # if dtime_units is not None:sta.attrs["dtime_units"] = dtime_units
    # if data_source is not None:sta.attrs["data_source"] = data_source
    # if level_type is not None: sta.attrs["data_type"] = level_type
    # if var_name is not None: sta.attrs["var_name"] = var_name
    # if var_cn_name is not None:sta.attrs["var_cn_name"] = var_cn_name
    # if var_units is not None:sta.attrs["var_units"] = var_units
    # if valid_time is not None:sta.attrs["valid_time"] = valid_time
    # if data_start_columns is not None:sta.attrs["data_start_columns"] = data_start_columns
    
    if units is not None: sta.attrs["units"] = units
    if model_var is not None: sta.attrs["model_var"] = model_var
    if dtime_units is not None: sta.attrs["dtime_units"] =dtime_units
    if level_type is not None: sta.attrs["level_type"] = level_type
    if time_type is not None: sta.attrs["time_type"] = time_type
    if time_bounds is not None: sta.attrs["time_bounds"] = time_bounds


def get_undim_data_names(sta):
    '''

    :param sta:
    :return:
    '''
    coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    columns = sta.columns
    data_columns = []
    for column in columns:
        if column not in coor_columns:
            if column.find("dim_type")!=0:
                data_columns.append(column)
    return data_columns

def get_stadata_names(sta):
    '''
    get_data_names() 获取站点数据的要素名
    :param sta: 站点数据
    :return: 要素名列表
    '''
    #coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    #columns = sta.columns
    #data_columns = []
    #for column in columns:
    #    if column not in coor_columns:
    #        data_columns.append(column)
    columns = sta.columns.values
    if "data_start_columns" not in sta.attrs.keys():
        sta.attrs["data_start_columns"] = 6
    data_start_column = sta.attrs["data_start_columns"]
    data_columns = columns[data_start_column:].tolist()
    return data_columns


def get_expanded_coord_names(sta):
    columns = sta.columns.values
    data_start_column = sta.attrs["data_start_columns"]
    ecc = columns[0:data_start_column].tolist()
    return ecc

def get_coord_names():
    '''

    :return: 站点数据基本信息列名['level', 'time', 'dtime', 'id', 'lon', 'lat']列表
    '''
    return ['level', 'time', 'dtime', 'id', 'lon', 'lat']

def set_stadata_names(sta,data_name_list):
    '''
    更改 要素名，和添加缺省列
    :param sta: 站点数据
    :param data_name: 站点数据 要素名
    :return: 更改要素名名后的站点数据
    '''
    if not isinstance(data_name_list,list):
        data_name_list = [data_name_list]
    #if isinstance(data_name_list,list):
    coor_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat']
    for data_name in data_name_list:
        coor_columns.append(data_name)
    sta.columns = coor_columns
    #else:
    #    print("输出名称设置不成功，数据名称列表参数需为list形式")
    return

def set_stadata_coords(sta,level = None,time = None,dtime = None,id = None,lat = None,lon = None):
    '''
    set_time_dtime_level_name 设置time_dtime_level 的值  并且设置要素名
    :param sta: 站点数据
    :param time: 起报时
    :param dtime: 时效
    :param level: 层次
    :param data_name: 要素名
    :return:  站点数据
    '''
    if time is not None:
        time1 = meteva_base.all_type_time_to_time64(time)
        sta.loc[:,'time'] = time1
        if len(sta.index) ==1:
            sta["time"] = sta["time"].astype(np.datetime64)
    if dtime is not None:
        sta.loc[:,'dtime'] = dtime
        dtime_type = str(sta["dtime"].dtype)
        if dtime_type.find("int") < 0:
            sta[["dtime"]] = sta[['dtime']].astype(int)
    if level is not None:
        sta.loc[:,'level'] = level
    if id is not None:
        sta.loc[:,"id"] = id
    if lat is not None:
        sta.loc[:,"lat"] = lat
    if lon is not None:
        sta.loc[:,"lon"] = lon



def reset_id(sta):
    '''
    输入的sta的站号中可能有些站号包含a-z,A-Z的字母，对此将这些字母转换为对应的ASCII数字，再将整个字符串格式的站号转换为数值形式
    返回sta站号为整型
    '''
    #print(sta)

    id_type = str(sta["id"].dtype)
    if id_type.find("int32") <0:
        values = sta['id'].values
        if id_type.find("int") >=0 or id_type.find("float") >=0:
            int_id = values.astype(np.int32)
            sta['id'] = int_id
        else:
            int_id = np.zeros(len(values))
            for i in range(len(values)):
                if isinstance(values[i],str):
                    strs = values[i]
                    if strs =="":
                        strs_int=999999
                    else:
                        strs_int = ""
                        for s in strs:
                            if s.isdigit():
                                strs_int += s
                            else:
                                strs_int += str(ord(s))
                    int_id[i] = int(strs_int)
                else:
                    int_id[i] = values[i]

            int_id = int_id.astype(np.int32)
            sta['id'] = int_id
    return

def get_stadata_attrs(sta,
              default_units='',
              default_model='',
              default_dtime_units='hour',
              default_level_type='isobaric',
              default_time_type='UT',
              default_time_bounds=[0,0]):
    try:
        if 'units' in sta.attrs:
            units=str(sta.attrs['units'])
        else:
            units=default_units
            
        if 'model' in sta.attrs:
            model=str(sta.attrs['model'])
        else:
            model=default_model
            
        if 'dtime_units' in sta.attrs:
            dtime_units=sta.attrs['dtime_units']
        else:
            dtime_units=str(default_dtime_units)
            
        if 'level_type' in sta.attrs:
            level_type=str(sta.attrs['level_type'])
        else:
            level_type=default_level_type
            
        if 'time_type' in sta.attrs:
            time_type=str(sta.attrs['time_type'])
        else:
            time_type=default_time_type
            
        if 'time_bounds' in sta.attrs:
            time_bounds=list(sta.attrs['time_bounds'])
        else:
            time_bounds=default_time_bounds
        
        return units,model,dtime_units,level_type,time_type,time_bounds
    
    except Exception as ex:
        raise ex

def set_stadata_coords_dtype(sta,
                            level_type=np.int16,
                            dtime_type=np.int16,
                            id_type=np.int32,
                            lat_type=np.float64,
                            lon_type=np.float64,
                            data_type = np.float32
    ):
    sta0=sta.copy()
    try:
        sta0['level']=sta0['level'].astype(level_type)
        sta0['time']=pd.to_datetime(sta0['time'])
        sta0['dtime']=sta0['dtime'].astype(dtime_type)
        sta0['id']=sta0['id'].astype(id_type)
        sta0['lon']=sta0['lon'].astype(lon_type)
        sta0['lat']=sta0['lat'].astype(lat_type)
        for column in sta0:
            if column  not in ['id','level','time','dtime','lon','lat']:
                sta0[column]=sta0[column].astype(data_type)
    except Exception as ex:
        print("ERROR in stadata dtype convert!")
        print(ex)
    return sta0



import datetime
import numpy as np
import re
import pandas as pd

#所有类型的时间转换为time64
def all_type_time_to_time64(time0):
    if isinstance(time0,np.datetime64):
        return time0
    elif isinstance(time0,datetime.datetime):
        return np.datetime64(time0)
    elif type(time0) == str:
        return str_to_time64(time0)
    elif isinstance(time0,int):
        time1 = datetime.datetime.utcfromtimestamp(time0 / 1000000000)
        np.datetime64(time1)
        return np.datetime64(time1)
    else:
        print("时间类型不识别")
        return None

def all_type_time_to_datetime(time0):
    if isinstance(time0,int):
        time1 = datetime.datetime.utcfromtimestamp(time0 / 1000000000)
        return time1
    time64 = all_type_time_to_time64(time0)
    time1 = time64.astype(datetime.datetime)
    if isinstance(time1,int):
        time1 = datetime.datetime.utcfromtimestamp(time1/1000000000)
    return time1

def all_type_time_to_str(time0):
    time1 = all_type_time_to_time64(time0)
    return time_to_str(time1)


#所有的timedelta类型的数据转为timedelta64类型的时间格式
def all_type_timedelta_to_timedelta64(timedelta0):
    if isinstance(timedelta0,np.timedelta64):
        return timedelta0
    elif isinstance(timedelta0,datetime.timedelta):
        return np.timedelta64(timedelta0)
    elif type(timedelta0) == str:
        timedelta1 = str_to_timedelta(timedelta0)
        return np.timedelta64(timedelta1)
    else:
        print("时效类型不识别")
        return None

def all_type_timedelta_to_timedelta(timedelta0):
    if isinstance(timedelta0,datetime.timedelta):
        return timedelta0
    elif isinstance(timedelta0,np.timedelta64):
        seconds = int(timedelta0 / np.timedelta64(1, 's'))
        timedelta1 = datetime.timedelta(seconds = seconds)
        return timedelta1
    elif type(timedelta0) == str:
        timedelta1 = str_to_timedelta(timedelta0)
        return timedelta1
    else:
        print("时效类型不识别")
        return None

#str类型的时间转换为timedelta64类型的时间
def str_to_timedelta(timedalta_str):
    num_str = ''.join([x for x in timedalta_str if x.isdigit()])
    num = int(num_str)
    # 提取出dtime_type类型
    TIME_type = re.findall(r"\D+", timedalta_str)[0].lower()
    if TIME_type == 'h':
        return datetime.timedelta(hours=num)
    elif TIME_type == 'd':
        return datetime.timedelta(days=num)
    elif TIME_type == 'm':
        return datetime.timedelta(minutes=num)
    else:
        print("输入的时效格式不识别")
        return None

#str类型的时间转换为time64类型的时间
def str_to_time64(time_str):
    str1 = ''.join([x for x in time_str if x.isdigit()])
    # 用户输入2019041910十位字符，后面补全加0000，为14位统一处理
    if len(str1) == 4:
        str1 += "0101000000"
    elif len(str1) == 6:
        str1 +="01000000"
    elif len(str1) == 8:
        str1 +="000000"
    elif len(str1) == 10:
        str1 +="0000"
    elif len(str1) == 12:
        str1 +="00"
    elif len(str1) == 14:
        str1 = time_str
    else:
        print("输入日期格式不识别，请检查！")

    # 统一将日期变为datetime类型
    time = datetime.datetime.strptime(str1, '%Y%m%d%H%M%S')
    time64 = np.datetime64(time)
    return time64



#datetime64类型的数据转换为str类型
def time_to_str(time):
    if isinstance(time,np.datetime64):
        str1 = str(time).replace("-", "").replace(" ", "").replace(":", "").replace("T", "")[0:14]
    else:
        str1 = time.strftime("%Y%m%d%H%M%S")
    return str1


#字符转换为datetime
def str_to_time(str0):
    num = ''.join([x for x in str0 if x.isdigit()])
    # 用户输入2019041910十位字符，后面补全加0000，为14位统一处理
    if len(num) == 4:
        num += "0101000000"
    elif len(num) == 6:
        num += "01000000"
    elif len(num) == 8:
        num += "000000"
    elif len(num) == 10:
        num += "0000"
    elif len(num) == 12:
        num += "00"
    elif len(num) > 12:
        num = num[0:12]
    else:
        print("输入日期有误，请检查！")
    # 统一将日期变为datetime类型
    return datetime.datetime.strptime(num, '%Y%m%d%H%M%S')


def get_dtime_of_path(path_model,path):
    ttt_index = path_model.find("TTT")
    if (ttt_index >= 0):
        ttt = int(path[ttt_index:ttt_index + 3])
    else:
        ttt = 0
    return ttt
def get_time_of_path(path_model,path):
    yy_index = path_model.find("YYYY")
    if  yy_index < 0:
        yy_index = path_model.find("YY")
        if(yy_index <0):
            yy = 2000
        else:
            yy = int(path[yy_index: yy_index + 2]) + 2000
    else:
        yy = int(path[yy_index: yy_index+4])

    mm_index = path_model.find("MM")
    if(mm_index >=0):
        mm = int(path[mm_index:mm_index+2])
    else:
        mm = 1

    dd_index = path_model.find("DD")
    if(dd_index>=0):
        dd = int(path[dd_index:dd_index + 2])
    else:
        dd = 1
    hh_index = path_model.find("HH")
    if(hh_index>=0):
        hh = int(path[hh_index:hh_index + 2])
    else:
        hh = 0
    ff_index = path_model.find("FF")
    if(ff_index>=0):
        ff = int(path[ff_index:ff_index + 2])
    else:
        ff = 0
    ss_index = path_model.find("SS")
    if(ss_index>=0):
        ss = int(path[ss_index:ss_index + 2])
    else:
        ss = 0
    return datetime.datetime(yy,mm,dd,hh,ff,ss)

#### 时间相关操作函数
def _get_date_from_str(input,get_datetime64=False):
    '''
    输入str/datetime64/datetime时，输出datetime或np.datetime（通过get_datetime64参数控制）
    '''
    if isinstance(input, int):##输入为数字时
        input = ''.join([x for x in gtime[i] if x.isdigit()])
    if type(input) == str:## 输入为字符串
        num = ''.join([x for x in input if x.isdigit()])
        # 用户输入2019041910十位字符，后面补全加0000，为14位统一处理
        if len(num) == 4:
            num += "0101000000"
        elif len(num) == 6:
            num +="01000000"
        elif len(num) == 8:
            num +="000000"
        elif len(num) == 10:
            num +="0000"
        elif len(num) == 12:
            num +="00"
        else:
            print("输入日期有误，请检查！")
        # 统一将日期变为datetime类型
        stime = datetime.datetime.strptime(num, '%Y%m%d%H%M%S')#返回datetime
        if get_datetime64 :
            stime = np.datetime64(stime)#返回datetime64                
    elif isinstance(input,np.datetime64):## 输入为datetime64
        if get_datetime64:
            stime = input
        else:
            stime = input.astype(datetime.datetime)
            print("here",stime)
            if isinstance(stime, int):
                stime = datetime.datetime.utcfromtimestamp(stime / 1000000000)                
    elif isinstance(input, datetime.datetime):## 输入为datetime.datetime
        if get_datetime64:
            stime = np.datetime64(input)#返回datetime64
        else:
            stime = input
    return(stime)

def get_date_list_create(gtime=['2020060700', '2020060712', 12], delta='H'):
    """
    ## 输入起始、终止、间隔时间，获取多时间列表
    gtime参数（start_date_str, end_date_str, hour_intervals）
    """
    num1 =[]
    if type(gtime[0]) == str:
        for i in range (0,2):
            num = ''.join([x for x in gtime[i] if x.isdigit()])
            num1.append(_get_date_from_str(num, get_datetime64=True))
        stime = num1[0]
        etime = num1[1]
    else:
        stime = _get_date_from_str(gtime[0])
        etime = _get_date_from_str(gtime[1])
    times = pd.date_range(stime, etime, freq=str(gtime[2])+delta)
    return(times.to_pydatetime())#返回datetime类型数组


def get_date_list_groupby(date_list, by='month'):
    """ 
    para:
        date_list   : 时间列表 List[datetime.datetime]
        by          : 分组方式，支持逐年'year'/逐月'month'/逐日'day'
    return：
        分组后的二维时间列表 List[List[datetime.datetime]]
        dim0为分组数，如逐年、逐月、逐日分组, 365天逐日分组时，len(dim0)=365
        dim1为具体每组的时间列表，如逐日分组时，列表中为同一天的具体时间datetime,  

    """
    time_index = pd.DatetimeIndex(date_list)
    time_index = time_index.sort_values(ascending=True)
    print(by)
    if by == 'year':
        temp = time_index.year*10000
    elif by == 'month':
        temp = time_index.year*10000 + time_index.month*100
    elif by == 'day':
        temp = time_index.year*10000 + time_index.month*100 + time_index.day
    else:
        raise ValueError('by参数应为： 逐年year/逐月month/逐日day， 请检查！！')
    time_pd = pd.DataFrame(temp, index=time_index,columns=['ymd'])
    time_pd_g = time_pd.groupby(by='ymd')
    time_monthly = []
    for f in time_pd_g:
        temp = f[1].index.to_pydatetime()
        time_monthly.append(temp)
    return time_monthly


#### 根据日期，时区，计算日出日落时间等
from math import sin, asin, cos, acos, tan, radians, pi, degrees
def _astro_time(date):
    Dn = int(date.strftime('%j'))
    # 第一步：计算太阳倾角(太阳直射点纬度)decl和equation of time
    gamma = 2*pi*(Dn - 1 + (12 - 12)/24)/365  #5.61757
    # 计算eqtime
    eqtime = 229.18*(0.000075 + 0.001868*cos(gamma) - 0.032077*sin(gamma) - 0.014615*cos(2*gamma) - 0.040849*sin(2*gamma))
    # 计算decl
    f1 = 0.006918
    f2 = 0.399912*cos(gamma)
    f3 = 0.070257*sin(gamma)
    f4 = 0.006758*cos(gamma*2)
    f5 = 0.000907*sin(gamma*2)
    f6 = 0.002697*cos(gamma*3)
    f7 = 0.001480*sin(gamma*3)
    decl = f1 - f2 + f3 - f4 + f5 - f6 + f7 #-0.3537
    return(eqtime, decl)

def get_sunrise_sunset_time(lon, lat, date, timezone=8):
    # lon lat 为经纬度，numpy数组形式,  date: 今天日期，datetime形式
        # 第零步：将日期转化为日期序数，1,2,3......，365；将时间转化为24小时制浮点数，如18:30转为18.5
    # return 日出时间及日落时间, 数组numpy，浮点数形式(非时分秒)

    if not isinstance(lon, np.ndarray):
        lon = np.array(lon)
    if not isinstance(lat, np.ndarray):
        lat = np.array(lat)
    if not isinstance(date, datetime.datetime):
        try:
            date = datetime.datetime.strptime(date, '%Y/%m/%d')
        except Exception as err:
            print(err)
            return None
    eqtime, decl = _astro_time(date)#天文时间相关
    # 第一步：计算太阳时角和方位角180度时的时间
    time_offset = eqtime + 4*lon - 60*timezone
    tst = 12*60 + time_offset
    ha = (tst/4 - 180)
    snoon = (720 - 4*lon - eqtime)/60 + timezone #——>ndarray

    lat = np.radians(lat)
    ha = np.radians(ha)  # 注意转为弧度

    # 第二步：计算日出日落时间
    zenith = radians(90.833)
    ha = np.arccos(cos(zenith)/(np.cos(lat)*cos(decl)) - np.tan(lat)*tan(decl))
    ha = np.degrees(ha)  # 注意转为角度 ,73.229 ->ndarray
    sunrise = snoon - ha/15
    sunset = snoon + ha/15
    return sunrise, sunset



if __name__ == "__main__":

    print(get_sunrise_sunset_time([116.5], [39.9], datetime.datetime(2023,12,22 ), 8)) # 北京14时
    



import pandas as pd
import numpy as np
import copy
import meteva
import datetime
import time


# 两个站点信息合并为一个，在原有的dataframe的基础上增加行数
def join(sta, sta1):
    if (sta is None):
        return sta1
    else:
        sta = pd.concat([sta, sta1])
    sta = sta.drop_duplicates(['id'])
    sta = sta.reset_index(drop=True)
    return sta


# 两个站点信息合并为一个，以站号为公共部分，在原有的dataframe的基础上增加列数
def merge(sta, sta1):
    if sta is None:
        return sta1
    elif sta1 is None:
        return sta
    else:
        df = pd.merge(sta, sta1, on='id', how='left')
        columns = list(sta.columns)
        len_sta = len(columns)
        # 删除合并后第二组时空坐标信息
        drop_col = list(df.columns[len_sta:len_sta + 6])
        df.drop(drop_col, axis=1, inplace=True)
        columns_dim = list(sta.columns)[0:7]
        columns_data = list(df.columns)[7:]
        columns = columns_dim + columns_data
        df.columns = columns
        return df


def that_the_name_exists(list, value):
    '''
    that_the_name_exists判断value是否在list中  如果存在改value直到不在list中为止
    :param list: 一个要素名列表
    :param value:  要素名
    :return:
    '''
    value = str(value)
    list = [str(i) for i in list]
    if value in list:
        value = str(value) + 'x'
        return that_the_name_exists(list, value)
    else:
        return value


#  两个站点信息合并为一个，以站号为公共部分，在原有的dataframe的基础上增加列数
def merge_on_all_dim(sta, sta1):
    '''
    merge_on_all_dim 合并两个sta_dataframe并且使要素名不重复
    :param sta: 一个站点dataframe
    :param sta1: 一个站点dataframe
    :return:
    '''
    if (sta is None):
        return sta1
    elif sta1 is None:
        return sta
    else:
        columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', 'alt']
        sta_value_columns = sta.iloc[:, 7:].columns.values.tolist()
        sta1_value_columns = sta1.iloc[:, 7:].columns.values.tolist()
        if len(sta_value_columns) >= len(sta1_value_columns):

            for sta1_value_column in sta1_value_columns:
                ago_name = copy.deepcopy(sta1_value_column)
                sta1_value_column = that_the_name_exists(sta_value_columns, sta1_value_column)
                sta1.rename(columns={ago_name: sta1_value_column},inplace=True)
        else:
            for sta_value_column in sta_value_columns:
                ago_name = copy.deepcopy(sta_value_column)
                sta_value_column = that_the_name_exists(sta1_value_columns, sta_value_column)
                sta.rename(columns={ago_name: sta_value_column})

        df = pd.merge(sta, sta1, on=columns, how='inner')
        return df


def merge_on_id_and_obTime(sta_list):
    '''
    merge_on_id_and_obTime  合并多个sta——dataframe  并且保证合并后的dataframe要素名不重复
    :param sta_list:   含有多个sta_dataframe的列表
    :return:
    '''
    intersection_of_data = None
    for sta in sta_list:
        sta['dtime'] = sta['dtime'].map(lambda x: datetime.timedelta(hours=x))
        sta['time'] = sta['time'] + sta['dtime']
        sta['dtime'] = 0

        intersection_of_data = meteva.base.function.put_into_sta_data.merge_on_all_dim(
            intersection_of_data, sta)
    return intersection_of_data
# def merge_on_all_dim(sta, sta1):
#     if (sta is None):
#         return sta1
#     elif sta1 is None:
#         return sta
#     else:
#         columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', 'alt']
#         df = pd.merge(sta, sta1, on=columns, how='inner')
#         return df
#
#
# def merge_on_id_and_obTime(sta_list):
#     intersection_of_data = None
#     for sta in sta_list:
#         sta['dtime'] = sta['dtime'].map(lambda x: datetime.timedelta(hours=x))
#         sta['time'] = sta['time'] + sta['dtime']
#         sta['dtime'] = 0
#         intersection_of_data = meteva.base.function.put_into_sta_data.merge_on_all_dim(
#             intersection_of_data, sta)
#     sta_value_columns_list = intersection_of_data.iloc[:, 7:].columns.values.tolist()
#     sta_value_columns_list = [str(sta_value) for sta_value in sta_value_columns_list]
#     sta_value_columns_set = set(sta_value_columns_list)
#
#     for sta_value in sta_value_columns_set:
#         num = sta_value_columns_list.count(sta_value)
#         if num > 1:
#             for i in range(num - 1):
#                 t = time.time()
#                 sta_value_columns_list[sta_value_columns_list.index(sta_value)] = sta_value + str(round(t * 1000))
#     sta_value_columns_list = ['level', 'time', 'dtime', 'id', 'lon', 'lat', 'alt'] + sta_value_columns_list
#     intersection_of_data.columns = sta_value_columns_list
#     return  intersection_of_data

    # lenght = len(sta_list)
    #     # corr_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', 'alt']
    #     # data = 'data'
    #     # for i in range(0, lenght):
    #     #     data1 = data + str(i)
    #     #     corr_columns.append(data1)
    #     # intersection_of_data.columns = corr_columns
    #     #
    #     # meteva.base.sta_data(intersection_of_data)
    #     # return intersection_of_data



import numpy as np
import pandas as pd
import xarray as xr

def put_into(grd_from, grd_to):
    '''
    两个格点网格信息的切片
    :param ged_from:源网格信息
    :param grd_to:目标网格信息
    :return:返回一个带有grd_to信息的多维xarray数据结构的网格信息
    比如grd_from 的网格范围是 10-20,110-150,
    grd_to的网格范围是 0-30,70-120，这时后者的网格范围就不能覆盖前者，
    如果后者范围是 0-30,70-150，则后者就能够覆盖前者。
    这是一比方，具体的问题包含6个维度，每个纬度都要判断一下后者的最大最小值是否能包含前者。
    如果不能包含就扩大到能包含，扩出来的部分先赋缺省值
    如果原来就能包含，就直接把那一块区域的值用grd_from代替
    根据grd_form中的坐标信息,判断grd_to 的坐标系能否覆盖前者
    如果能：
    把grd_from 中的数据覆盖掉grd_to中相同的网格部分
    如果不能：
    在grd_to中将坐标范围扩展成能覆盖前者，扩展出来的网格区域先设置为9999
    再将grd_from中的值覆盖掉grd_to中相同的网格部分
    '''
    #grd_from 的网格范围信息
    gf = grd_from.to_dataframe(name="")
    fslat = float(gf.index.get_level_values(4)[0])
    felat = float(gf.index.get_level_values(4)[-1])
    fslon = float(gf.index.get_level_values(5)[0])
    felon = float(gf.index.get_level_values(5)[-1])
    # grd_to 的网格范围信息
    gt = grd_to.to_dataframe(name="")
    tslat = float(gt.index.get_level_values(4)[0])
    tslat2 = float(gt.index.get_level_values(4)[1])
    telat = float(gt.index.get_level_values(4)[-1])
    tslon = float(gt.index.get_level_values(5)[0])
    tslon2 = float(gt.index.get_level_values(5)[1])
    telon = float(gt.index.get_level_values(5)[-1])
    tdlat = tslat2 - tslat
    tdlon = tslon2 - tslon

    #判断grd_from是否包含在grd_to范围内
    if fslat >= tslat and felat <=telat and fslon >= tslon and felon <= telon:
        gt_data = gt.iloc[fslat-tslat:felat-tslat, fslon-tslon:felon-tslon] = gf.values
    else:
        #grd_from的纬度起始范围在grd_to的范围内不处理，不在就扩充到包含这个范围内。
        if fslat >= tslat:
            if felat <= telat:
                pass
            else:
                telat = felat
        else:
            tslat = fslat
        # grd_from的经度起始范围在grd_to的范围内不处理，不在就扩充到包含这个范围内。
        if fslon >= tslon:
            if felon <= telon:
                pass
            else:
                telon = felon
        else:
            tslon = fslon
        #经过转换之后，grd_to变成了一个扩展之后的最大的一个网格。
        gt_data = pd.DataFrame(columns=np.arange(tslon, telon+tdlon, tdlon), index=np.arange(tslat, telat+tdlon, tdlat))
        print(gt_data.values)
        gt_data = gt_data.iloc[fslat - tslat:felat - tslat, fslon - tslon:felon - tslon] = gf.values
        gt_data.fillna(9999, inplace=True)
    tnlon = (telon - tslon)/tdlon + 1
    tnlat = (telat - tslat) /tdlat + 1
    lon = np.arange(tnlon) * tdlon + tslon
    lat = np.arange(tnlat) * tdlat + tslat
    fmember_list = list(gf.index.get_level_values(0).unique())
    flevel_list = list(gf.index.get_level_values(1).unique())
    ftime_list = list(gf.index.get_level_values(2).unique())
    fdt_list = list(gf.index.get_level_values(3).unique())
    tmember_list = list(gt.index.get_level_values(0).unique())
    tlevel_list = list(gt.index.get_level_values(1).unique())
    ttime_list = list(gt.index.get_level_values(2).unique())
    tdt_list = list(gt.index.get_level_values(3).unique())
    new_member_list =tmember_list
    new_level_list = tlevel_list
    new_time_list = ttime_list
    new_dt_list = tdt_list
    for i in fmember_list:
        if i not in tmember_list:
            new_member_list.append(i)
    for i in flevel_list:
        if i not in tlevel_list:
            new_level_list.append(i)
    for i in ftime_list:
        if i not in ttime_list:
            new_time_list.append(i)
    for i in fdt_list:
        if i not in tdt_list:
            new_dt_list.append(i)
    data = gt_data.reshape(len(new_member_list), len(new_level_list), len(new_time_list), len(new_dt_list), tnlat, tnlon)
    return (xr.DataArray(data, coords={'member': new_member_list, 'level': new_level_list, 'time': new_time_list, 'dt': new_dt_list,
                                       'lat': lat, 'lon': lon},
                         dims=['member', 'level', 'time', 'dt', 'lat', 'lon']))



import meteva
import numpy as np
import pandas as pd

#格点转换为站点
def transform(grd):
    x = grd['lon']
    y = grd['lat']
    grid_x,grid_y = np.meshgrid(x,y)
    grid_num = len(x) * len(y)
    dat = np.empty((grid_num,4))
    dat[:,0] = np.arange(grid_num)
    dat[:,1] = grid_x.reshape(-1)
    dat[:,2] = grid_y.reshape(-1)
    dat[:,3] = grd.values.reshape(-1)
    member = grd['member'][0]
    df = pd.DataFrame(dat,columns=['id','lon','lat',member])
    sta = meteva.base.basicdata.sta_data(df)
    grid = meteva.base.basicdata.get_grid_of_data(grd)
    sta['time'] = grid.stime
    sta['dtime'] = grid.sdtimedelta
    sta['level'] = grid.levels[0]
    sta['alt'] = 9999
    return sta

#格点到站点的插值
def interpolation_nearest(grd,sta,other_info='left'):
    grid = meteva.base.basicdata.get_grid_of_data(grd)
    sta1 = meteva.base.function.get_from_sta_data.sta_in_grid_xy(sta, grid)
    dat0 = grd.values
    dat = np.squeeze(dat0)

    ig = np.round((sta1['lon'] - grid.slon) // grid.dlon).astype(dtype = 'int16')
    jg = np.round((sta1['lat'] - grid.slat) // grid.dlat).astype(dtype = 'int16')

    dat_sta= dat[jg,ig]
    data_name = meteva.base.basicdata.get_data_names(sta)[0]
    sta1.loc[:, data_name] = dat_sta[:]
    if other_info == 'left':
        sta1['time'] = grid.stime
        sta1['dtime'] = grid.dtimes[0]
        sta1['level'] = grid.levels[0]
        meteva.base.basicdata.set_data_names(sta1,grid.members)
    return sta1

#双线性格点到站点的插值
def interpolation_linear(grd,sta,other_info='left'):
    grid = meteva.base.basicdata.get_grid_of_data(grd)
    sta1 = meteva.base.function.get_from_sta_data.sta_in_grid_xy(sta, grid)
    dat0 = grd.values
    dat = np.squeeze(dat0)

    ig = ((sta1['lon'] - grid.slon) // grid.dlon).astype(dtype = 'int16')
    jg = ((sta1['lat'] - grid.slat) // grid.dlat).astype(dtype = 'int16')
    dx = (sta1['lon'] - grid.slon) / grid.dlon - ig
    dy = (sta1['lat'] - grid.slat) / grid.dlat - jg
    c00 = (1 - dx) * (1 - dy)
    c01 = dx * (1 - dy)
    c10 = (1-dx) * dy
    c11 = dx * dy

    ig1 = np.minimum(ig + 1, grid.nlon - 1)
    jg1 = np.minimum(jg + 1, grid.nlat - 1)
    dat_sta= c00 * dat[jg,ig] + c01 * dat[jg,ig1] + c10 * dat[jg1,ig] + c11 * dat[jg1,ig1]
    data_name = meteva.base.get_stadata_names(sta)[0]
    sta1.loc[:, data_name] = dat_sta[:]
    if other_info == 'left':
        sta1['time'] = grid.stime
        sta1['dtime'] = grid.dtimes[0]
        sta1['level'] = grid.levels[0]
        meteva.base.set_stadata_names(sta1,grid.members)
    return sta1


#三维格点到站点的插值
def interpolation_cubic(grd,sta,other_info = 'left'):
    grid = meteva.base.basicdata.get_grid_of_data(grd)
    sta1 = meteva.base.function.get_from_sta_data.sta_in_grid_xy(sta, grid)
    dat0 = grd.values
    dat = np.squeeze(dat0)
    ig = ((sta1['lon'] - grid.slon) // grid.dlon).astype(dtype = 'int16')
    jg = ((sta1['lat'] - grid.slat) // grid.dlat).astype(dtype = 'int16')
    dx = (sta1['lon'] - grid.slon) / grid.dlon - ig
    dy = (sta1['lat'] - grid.slat) / grid.dlat - jg
    data_name = meteva.base.basicdata.get_data_names(sta1)[0]
    for p in range(-1,3,1):
        iip = np.minimum(np.maximum(ig+p,0),grid.nlon-1)
        fdx = cubic_f(p, dx)
        for q in range(-1,3,1):
            jjq = np.minimum(np.maximum(jg+q,0),grid.nlat-1)
            fdy = cubic_f(q,dy)
            fdxy = fdx * fdy
            sta1[data_name] +=  fdxy * dat[jjq,iip]
    if other_info == 'left':
        sta1['time'] = grid.stime
        sta1['dtime'] = grid.dtimes[0]
        sta1['level'] = grid.levels[0]
        meteva.base.basicdata.set_data_names(sta1,grid.members)
    return sta1

def cubic_f(n, dx):
    if (n == -1):
        return -dx * (dx - 1) * (dx - 2) / 6
    elif (n == 0):
        return (dx + 1) * (dx - 1) * (dx - 2) / 2
    elif (n == 1):
        return -(dx + 1) * dx * (dx - 2) / 2
    else:
        return (dx + 1) * dx * (dx - 1) / 6



import  numpy as np
import pandas as pd
def get_from(grd_from, grd_to):
    '''
    格点网格的切分、分片
    :param grd_from:源网格信息
    :param grd_to:需要映射的目标网格信息
    :return:两个网格的重合部分，并赋值返回。
    # 首先根据grid_to定一个初始的网格场grd_to，并且将取值都设为缺省
    # 然后从grd_from里面找到两个网格重合的区域的那部分取值，将其赋值到grd_to
    
    '''
    gf = grd_from.to_dataframe(name="")
    fslat = float(gf.index.get_level_values(4)[0])
    felat = float(gf.index.get_level_values(4)[-1])
    fslon = float(gf.index.get_level_values(5)[0])
    felon = float(gf.index.get_level_values(5)[-1])
    gt = grd_to.to_dataframe(name="")
    tslat = float(gt.index.get_level_values(4)[0])
    tslat2 = float(gt.index.get_level_values(4)[1])
    telat = float(gt.index.get_level_values(4)[-1])
    tslon = float(gt.index.get_level_values(5)[0])
    tslon2 = float(gt.index.get_level_values(5)[1])
    telon = float(gt.index.get_level_values(5)[-1])
    #直接相减两个相邻的数为零，需要相隔lo的长度个数才能求出经度
    tdlat = tslat2 - tslat
    tdlon = tslon2 - tslon
    #根据grid_to定一个初始的网格场grd_to，并且将取值都设为缺省
    gt_data = pd.DataFrame(columns=np.arange(tslon, telon + tdlon, tdlon), index=np.arange(tslat, telat + tdlon, tdlat))
    print(gt_data.values)
    #求出两个网格的公共区域
    slon = np.max([int(fslon * 100), int(tslon * 100)]) / 100
    elon = np.min([int(felon * 100), int(telon * 100)]) / 100
    slat = np.max([int(fslat * 100), int(tslat * 100)]) / 100
    elat = np.min([int(felat * 100), int(telat * 100)]) / 100
    print("起始范围：{}--{} {}--{}".format(slon, elon, slat, elat))
    if slon >=elon or slat >=elat:
        print("两者不存在公共区域！")
        return 0
    else:
        data = grd_from.values()
        dat = np.squeeze(data)
        print(dat.shape)
        #切分出两个网格公共区域部分的值
        fill_data = dat[fslat - slat:felat - slat,fslon - slon:felon -slon ]
        #从grd_from中求出公共区域的值，并进行填充
        gt_data = gt_data.iloc[fslat - tslat:felat - tslat, fslon - tslon:felon - tslon] = fill_data
        gt_data.fillna(9999, inplace=True)
        grd_to.values = gt_data
        return grd_to



import math
import datetime
import re
from copy import deepcopy
import numpy as np
import pandas as pd

import meteva_base


class grid:
    '''
        定义一个格点的类grid，来存储网格的范围包括（起始经纬度、格距、起止时间，时间间隔，起止时效，时效间隔，层次列表，数据成员）
        约定坐标顺序为: member, time,ddtime, level, lat,lon
    '''
    def __init__(self,glon, glat, gtime=None, dtime_list=None, level_list=None, member_list = None,
                        units_attr      = '',#数据单位
                        model_var_attr  = '',#补充模式/要素信息
                        dtime_units_attr= 'hour',# hour/minite
                        level_type_attr = 'isobaric',# isobaric/attitude
                        time_type_attr  = 'UT',#UT/BT
                        time_bounds_attr= [0,0],#数据起止时间
                        ):
        """
        ## 格点坐标设置
        param glon : 东西向网格信息，list列表，[最小，最大，间隔]
        param glat : 南北向网格信息，list列表，[最小，最大，间隔]
        param gtime: 起报时间网格信息：若gtime只包含1个元 素，其内容为datetime变量，或者常用字符形式时间，例如"2019123108"、"2019-12-31:08"或"19年12月31日08时"等方式都能兼容。
                    若gtime包含三个元素，其内容为[起始时间,结束时间,时间间隔], 其中时间格式可以是datetime变量或者字符串形式，时间间隔为字符串，例如"12h"代表12小时，"30m"代表30分钟。
                    若为列表且非上述要素，则是直接的datetime列表
        param dtime_list: 时效信息，list列表，其元素dtime为整数, 默认为[0]
        param level_list: 层次名称信息，list列表，默认为[0]
        param member_list:成员名称信息，list列表，默认为['data0']
        ## 格点属性设置
        param units_attr      : 数据单位, 默认为None,
        param model_var_attr  : 补充模式/要素信息,默认为None
        param dtime_units_attr: 预报时效单位，hour/minite,默认为'hour' 
        param level_type_attr : 层次单位，isobaric/attitude，默认为'isobaric'等压面
        param time_type_attr  : 起报时间类型，UT（世界时间）/BT(北京时间)，默认'UT'
        time_bounds_attr      : 数据起止时间，如24h降水为[-24,0], 默认[0,0]
        
        """
        #提取成员维度信息
        if(member_list is None):
            self.members =['data0']
        else:
            self.members = member_list
        ############################################################################
        #提取层次维度信息
        if(level_list is None):
            self.levels =[0]
        else:
            self.levels = level_list
        ############################################################################
        #提取时间维度信息
        # try:
        if gtime is None:
            self.stime = np.datetime64('2020-01-01T00:00:00.000000')
            self.etime = np.datetime64('2020-01-01T00:00:00.000000')
            self.dtime_int = 1
            self.dtime_type = "h"
            self.gtime = pd.date_range(self.stime, self.etime, freq=str(self.dtime_int)+self.dtime_type ).to_pydatetime()#[datetime]
        else:
            try:
                _ = len(gtime)
            except:
                gtime = [gtime]
            if len(gtime) == 1:
                self.gtime = meteva_base.basicdata.utils.get_time_input_single(gtime)

            elif len(gtime) ==3 and isinstance(gtime[2],str) and len(gtime[2])<=5:
                self.gtime = meteva_base.basicdata.utils.get_time_input_three(gtime)

            else:
                if isinstance(gtime[0], datetime.datetime):
                    self.gtime = gtime
                else:
                    self.gtime = [meteva_base.all_type_time_to_datetime(dt) for dt in gtime]#[datetime]
        self.stime_str = self.gtime[0].strftime("%y%m%d%H%M")
        # except Exception as err:
            # print('gtime ERROR: ',err)
        ############################################################################
        #提取预报时效维度信息
        if dtime_list is None:
            self.dtimes = [0]
        else:
            self.dtimes = dtime_list
        ############################################################################
        #提取经度信息

        self.slon = get_true_value(glon[0])
        self.elon = get_true_value(glon[1])
        self.dlon = get_true_value(glon[2])
        nlon = 1 + (self.elon - self.slon) / self.dlon
        error = abs(round(nlon) - nlon)/nlon
        if (error > 0.01):
            self.nlon = int(math.ceil(nlon))
        else:
            self.nlon = int(round(nlon))

        self.elon = get_true_value(self.slon + (nlon - 1) * self.dlon)
        self.glon = [self.slon,self.elon,self.dlon]
        ############################################################################
        #提取纬度信息
        self.slat = get_true_value(glat[0])
        self.elat = get_true_value(glat[1])
        self.dlat = get_true_value(glat[2])
        nlat = 1 + (self.elat - self.slat) / self.dlat
        error = abs(round(nlat) - nlat)/nlat
        if (error > 0.01):
            self.nlat = int(math.ceil(nlat))
        else:
            self.nlat = int(round(nlat))
        self.elat = get_true_value(self.slat + (nlat - 1) * self.dlat)
        self.glat = [self.slat,self.elat,self.dlat]
        ## 格点数据属性设置(6属性)
        self.units       = units_attr
        self.model_var   = model_var_attr
        self.dtime_units = dtime_units_attr
        self.level_type  = level_type_attr
        self.time_type   = time_type_attr
        self.time_bounds = time_bounds_attr
        return




    #对原有的格点数据进行一次深拷贝，不改变原有的值和结构。
    def copy(self):
        return deepcopy(self)

    #reset的作用是把网格的坐标间隔统一为正数。
    def reset(self):
        if (self.dlon > 0 and self.dlat > 0):
            pass
        if (self.dlat < 0):
            tran = self.slat
            self.slat = self.elat
            self.elat = tran
            self.dlat = abs(self.dlat)
        if (self.dlon < 0):
            tran = self.slon
            self.slon = self.elon
            self.elon = tran
            self.dlon = abs(self.dlon)
        return


    def __str__(self):
        '''
        重置系统自动的函数，在print(grid) 的时候可以很整齐的看到所有信息
        :return:  string
        '''
        grid_str = ""
        grid_str += "## grid_coordinates ##" + "\n"
        grid_str += "members:" + str(self.members) +"\n"
        grid_str += "levels:" + str(self.levels) + "\n"
        # grid_str += "gtime:" + str([self.stime_str,self.etime_str,self.dtime_str]) + "\n"
        grid_str += "gtime:" + str(self.stime_str)  +"\n"
        grid_str += "dtimes:" + str(self.dtimes)  +"\n"
        grid_str += "glon:" + str(self.glon) + "\n"
        grid_str += "glat:" + str(self.glat) + "\n"
        grid_str += "## grid_attributes ##" +  "\n"
        grid_str += "units:" + str(self.units) + "\n"
        grid_str += "model_var:" + str(self.model_var) + "\n"
        grid_str += "dtime_units:" + str(self.dtime_units) + "\n"
        grid_str += "level_type:" + str(self.level_type) + "\n"
        grid_str += "time_type:" + str(self.time_type) + "\n"
        grid_str += "time_bounds:" + str(self.time_bounds) + "\n"
        return grid_str

def get_true_value(value):
    dlon2 = round(value, 2)
    dlon3 = round(value, 3)
    dlon4 = round(value, 4)
    dlon5 = round(value, 5)
    if dlon2 == dlon3 and dlon3 == dlon4:
        return dlon3
    elif dlon3 == dlon4 and dlon5 == dlon4:
        return dlon3
    else:
        return value


def get_grid_of_data(grid_data0):
    '''
     获取grid的数据values值
    :param grid_data0:初始化之后的网格数据
    :return:返回grid数据。
    '''
    member_list = grid_data0['member'].values
    level_list = grid_data0['level'].values
    times = grid_data0['time'].values
    gtime = times

    gdt = grid_data0['dtime'].values.tolist()
    attrs_name = list(grid_data0.attrs)


    lons = grid_data0['lon'].values
    dlon = get_true_value(lons[1] - lons[0])
    glon = [get_true_value(lons[0]), get_true_value(lons[-1]), dlon]
    lats = grid_data0['lat'].values
    dlat = get_true_value(lats[1] - lats[0])
    glat = [get_true_value(lats[0]), get_true_value(lats[-1]), dlat]

    # units_attr       = grid_data0.attrs['units']
    # model_var_attr   = grid_data0.attrs['model_var']
    # dtime_units_attr = grid_data0.attrs['dtime_units']
    # level_type_attr  = grid_data0.attrs['level_type']
    # time_type_attr   = grid_data0.attrs['time_type']
    # time_bounds_attr = grid_data0.attrs['time_bounds']
    units,model,dtime_units,level_type,time_type,time_bounds = meteva_base.basicdata.get_griddata_attrs(grid_data0)

    grid01 = grid(glon, glat, gtime, gdt, level_list, member_list,
                        units_attr      = units,
                        model_var_attr  = model,
                        dtime_units_attr= dtime_units,
                        level_type_attr = level_type,
                        time_type_attr  = time_type,
                        time_bounds_attr= time_bounds)
    return grid01


def reset_grid(grid0):
    if grid0.dlat <0:
        grid0.dlat = - grid0.dlat
        tran = grid0.slat
        grid0.slat = grid0.elat
        grid0.elat = tran
    if grid0.dlon <0:
        grid0.dlon = - grid0.dlon
        tran = grid0.slon
        grid0.slon = grid0.elon
        grid0.elon = tran
    return




import numpy as np
import meteva

#获取num层的网格数据
def get_member(grd,num):
    dat = grd.values[num,0,0,0,:,:]
    grid0 = meteva.base.basicdata.get_grid_of_data(grd)
    grid1 = meteva.base.basicdata.grid(grid0.glon,grid0.glat,grid0.gtime,grid0.dtimes,grid0.levels)
    grd1= meteva.base.basicdata.grid_data(grid1,dat)
    return grd1

#获取网格数据的平均值
def get_mean(grd,member_name):
    grid0 = meteva.base.basicdata.get_grid_of_data(grd)
    grid1 = meteva.base.basicdata.grid(grid0.glon,grid0.glat,grid0.gtime,grid0.dtimes,grid0.levels,members=[member_name])
    dat = np.squeeze(grd.values)
    dat = np.mean(dat,axis = 0)
    grd1 = meteva.base.basicdata.grid_data(grid1,dat)
    return grd



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Filename: CMADaasAccess.py

'''
Created on Sep 15, 2020

@author: anduin
'''
import meteva_base
import datetime
import json
from meteva_base.io.loglib import *
from meteva_base.io.httpconn import HttpConn
from meteva_base.io.SignGenUtil import SignGenUtil
import numpy as np

class CMADaasAccess():
    """description of class
    cmadaas data access
    """
    @staticmethod
    def get_times(dts):
        try:
            if type(dts) is list:
                return ','.join(dt.strftime('%Y%m%d%H%M%S') for dt in dts)
            else:
                return dts.strftime('%Y%m%d%H%M%S')
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess get_times except %s %s' % (str(dts), str(data)))
            raise data
            # return None

    @staticmethod
    def set_times_in_params(dts, params, deep_copy=True):
        try:
            rst = params
            if deep_copy:
                import copy
                rst = copy.deepcopy(params)

            rst['times'] = CMADaasAccess.get_times(dts)

            return rst
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess set_times_in_params except %s %s' % (str(dts), str(data)))
            raise data
            # return None

    @staticmethod
    def get_timerange(dts):
        try:
            rst = ''
            if type(dts) is list:
                rst = ','.join(dt.strftime('%Y%m%d%H%M%S') for dt in dts)
            else:
                rst = dts.strftime('%Y%m%d%H%M%S')

            return '[' + rst + ']'
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess get_timerange except %s %s' % (str(dts), str(data)))
            raise data
            # return None

    @staticmethod
    def set_timerange_in_params(dts, params, deep_copy=True):
        try:
            rst = params
            if deep_copy:
                import copy
                rst = copy.deepcopy(params)

            rst['timeRange'] = CMADaasAccess.get_timerange(dts)

            return rst
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess set_timerange_in_params except %s %s' % (str(dts), str(data)))
            raise data
            # return None

    @staticmethod
    def get_query(params):
        '''
        生成url中query部分
        '''
        try:
            querystr = ''
            for key, item in params.items():
                querystr += key + "=" + str(item) + "&"

            if (querystr):
                querystr = querystr[:-1]

            return querystr
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess get_query except %s %s' % (str(params), str(data)))
            raise data
            # return None

    @staticmethod
    def get_url(url, qparams):
        '''
        生成url
        '''
        # import copy
        try:
            # signParams = copy.deepcopy(qparams)
            signParams = qparams
            signString = SignGenUtil.getSign(signParams)
            if signString is None:
                LogLib.error('CMADaasAccess getUrl getSign error %s %s' % (url, str(qparams)))
                return None

            signParams['sign'] = signString
            signParams.pop('pwd')

            querystr = CMADaasAccess.get_query(signParams)

            return url + querystr
        except Exception as data:
            # print(str(data))
            LogLib.error('CMADaasAccess getUrl except %s %s %s' % (url, str(qparams), str(data)))
            raise data
            # return None

    @staticmethod
    def get_default_headers():
        return {'Accept': 'application/json, text/plain, */*',
                'Accept-Encoding': 'gzip, deflate, br',
                'Accept-Language': 'zh-CN,zh;q=0.9',
                'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.100 Safari/537.36'}

    # only json is supported
    # 从cmadaas获得json格式数据，使用json库解析，构建一个dataframe返回。
    @staticmethod
    def read_data_from_cmadaas(fullurl):
        try:
            LogLib.info('CMADaasAccess read_data_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_data_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None

            jsonobj = json.loads(rstcontent)
            if jsonobj is None:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json loads error %s %s' % (rstcontent, str(fullurl)))
                return None

            if 'returnCode' not in jsonobj:
                LogLib.error(
                    'CMADaasAccess read_data_from_cmadaas json no returnCode %s %s' % (rstcontent, str(fullurl)))
                return None

            if jsonobj['returnCode'] != '0':
                if jsonobj['returnCode'] == '-1':
                    LogLib.info('CMADaasAccess read_data_from_cmadaas json no data %s %s' % (rstcontent, str(fullurl)))
                    return []
                else:
                    LogLib.error('CMADaasAccess read_data_from_cmadaas json no returnCode or returnCode != 0 %s %s' % (
                    rstcontent, str(fullurl)))
                    return None

            return jsonobj['DS']
        except Exception as data:
            LogLib.error('CMADaasAccess read_data_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None

    @staticmethod
    def read_griddata_from_cmadaas(fullurl, time=None, dtime=None):
        try:
            LogLib.info('CMADaasAccess read_data_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_data_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None

            jsonobj = json.loads(rstcontent)
            if jsonobj is None:
                LogLib.error('CMADaasAccess read_data_from_cmadaas json loads error %s %s' % (rstcontent, str(fullurl)))
                return None

            if 'returnCode' not in jsonobj:
                LogLib.error(
                    'CMADaasAccess read_data_from_cmadaas json no returnCode %s %s' % (rstcontent, str(fullurl)))
                return None

            griddata = CMADaasAccess.griddata_from_cmadaas_json(jsonobj, time=time, dtime=dtime)
            return griddata
        except Exception as data:
            LogLib.error('CMADaasAccess read_data_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None

    @staticmethod
    def griddata_from_cmadaas_json(json, time=None, dtime=None):
        try:
            lonS = round(json['startLon'], 4)
            # lonE = round(json['endLon'],3)
            latS = round(json['startLat'], 4)
            # latE = round(json['endLat'],3)
            dlon = round(json['lonStep'], 4)
            dlat = round(json['latStep'], 4)
            nlon = int(json['lonCount'])
            nlat = int(json['latCount'])
            lonE = lonS + (nlon - 1) * dlon
            latE = latS + (nlat - 1) * dlat
            # print(lonS,dlon,nlon,lonE)
            data = np.array(json['DS'])

            # data
            # print(data.shape)

            if time is not None:
                if not isinstance(time, list):
                    time = [time]  # 确保time为列表
            if dtime is not None:
                if not isinstance(dtime, list):
                    dtime = [dtime]  # 确保dtime为列表
            # print(time,dtime)
            grid_info = meteva_base.grid([lonS, lonE, dlon], [latS, latE, dlat]
                                         , gtime=time
                                         , dtime_list=dtime)
            # print(grid_info)
            grid = meteva_base.grid_data(grid_info, data)
        except Exception as data:
            print(json)
            return None
        return grid

    @staticmethod
    def dataframe_from_dict_json(j_dict, default=np.nan):
        import pandas as pd
        try:
            LogLib.info('CMADaasAccess Convert dataFrame from JSON_dict {0}'.format(
                datetime.datetime.now().strftime("%Y%m%d%H%M")))
            pddata = pd.DataFrame.from_dict(j_dict)
            ## 更改dataframe
            sta = meteva_base.sta_data(pddata, columns=['time', 'id', 'lat', 'lon', 'data0'])
            sta.level = 0
            sta.dtime = 0
            sta.time = pd.to_datetime(sta.time)
            sta.level = sta.level.astype(np.int16)
            sta.id = sta.id.astype(np.int32)
            sta.dtime = sta.dtime.astype(np.int16)
            sta.lon = sta.lon.astype(np.float32)
            sta.lat = sta.lat.astype(np.float32)
            try:
                sta.data0 = sta.data0.astype(np.float32)
                sta.data0[sta.data0 >= 990000] = default
            except:
                pass
            return (sta)
        except Exception as data:
            LogLib.error('CMADaasAccess Convert dataFrame from JSON_dict {0}'.format(
                datetime.datetime.now().strftime("%Y%m%d%H%M")))
            return (None)

    # 从cmadaas获得文件数据。
    @staticmethod
    def read_file_from_cmadaas(fullurl):
        try:
            LogLib.info('CMADaasAccess read_file_from_cmadaas start %s' % (str(fullurl)))
            headers = CMADaasAccess.get_default_headers()
            host, url, ishttp = HttpConn.parseUrl(fullurl)
            rststatus, rstcontent, rstheaders = HttpConn.getWebPage(host, url, headers=headers, ishttp=ishttp)
            if rststatus != 200:
                LogLib.error('CMADaasAccess read_file_from_cmadaas net error %d %s' % (rststatus, str(fullurl)))
                return None
            return (rstcontent, rstheaders)
        except Exception as data:
            LogLib.error('CMADaasAccess read_file_from_cmadaas except %s %s' % (str(fullurl), str(data)))
            return None

    @staticmethod
    def write_data_to_file(save_path, grd, creat_dir=True):
        import os
        try:
            LogLib.info('FixWriteData write_data_to_file start %s' % (save_path))
            if creat_dir:
                path, file = os.path.split(save_path)
                if not os.path.exists(path):
                    os.makedirs(path)
            with open(save_path, 'wb') as f:
                f.write(grd)
            LogLib.info('FixWriteData write_data_to_file over %s' % (save_path))
            return True
        except Exception as data:
            LogLib.error('FixWriteData write_data_to_file except %s' % (str(save_path)))
            raise data

    ## 生成datetime类
    def get_datetime_from_str(dt='202011110200'):  # dt可以为任意长度字符串，也可为datetime类
        if type(dt) == str:
            # 用户输入2019041910十位字符，后面补全加0000，为14位统一处理
            if len(dt) == 4:
                num1 = dt + "0101000000"
            elif len(dt) == 6:
                num1 = dt + "01000000"
            elif len(dt) == 8:
                num1 = dt + "000000"
            elif len(dt) == 10:
                num1 = dt + "0000"
            elif len(dt) == 12:
                num1 = dt + "00"
            elif len(dt) == 14:
                num1 = dt
            else:
                print("输入日期有误，请检查！")
            # 统一将日期变为datetime类型
            time = datetime.datetime.strptime(num1, '%Y%m%d%H%M%S')
        else:
            time = dt
        return (time)

    ## 根据参数字典，生成REST格式url。
    @staticmethod
    def combine_url_from_para(qprams, time=datetime.datetime(2020, 11, 27, 0, 0), url=None,
                              SN_id="NMIC_MUSIC_CMADAAS", dataFormat='json', userId=None, pwd=None,
                              defaultNan=9999, show_url=False, time_name='times'):
        """
        dt： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据
        userID: 业务账号，必须填
        pwd： 业务账户密码，必须填
        dataFormat: 返回数据格式，默认为json
        show_url:  是否完整显示url，默认为不显示
        url: cmadaas的url前面固定部分，默认'http://10.40.17.54/music-ws/api?'
        SN_id: cmadaas服务节点代码，默认'NMIC_MUSIC_CMADAAS'
        time_name:时间参数的字典名(time或者times)
        """
        import copy
        params = copy.copy(qprams)
        LogLib.info('OBS Download from CmaDaas: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))

        ## 账户处理
        if userId is None:
            userId = meteva_base.cmadaas_set[2]
            LogLib.warning(
                'NO userID input,Using default USER: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        if pwd is None:
            pwd = meteva_base.cmadaas_set[3]
            LogLib.warning(
                'NO Password input,Using default pwd: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        ## 添加字典变量要素
        if "timeRange" in params:
            params['timeRange'] = (str(params['timeRange']).replace(' ', ''))
        if "userId" not in params:
            params["userId"] = userId
        #            params = params.pop("userId")
        if "pwd" not in params:
            params["pwd"] = pwd
        if not (("times" in params) or ("time" in params) or ("timeRange" in params)):
            params[time_name] = CMADaasAccess.get_times(time)
        if "serviceNodeId" not in params: params["serviceNodeId"] = SN_id
        if "timestamp" not in params: params["timestamp"] = str(int(datetime.datetime.now().timestamp() * 1000))
        if "nonce" not in params: params["nonce"] = '3696663f-d202-4570-b39d-16306f419575'
        if "dataFormat" not in params: params['dataFormat'] = dataFormat
        ## 生成url

        # print(params['timeRange'])
        if url is None:
            url = "http://" + meteva_base.cmadaas_set[0] + "/music-ws/api?"

        urlstr = CMADaasAccess.get_url(url, params)
        if show_url is True: print(urlstr)
        ## 返回url
        return (urlstr)

    @staticmethod
    def get_obs_micaps3_from_cmadaas(qparams, time="202011110200", save_path=None, filename=None, defaultNan=9999,
                                     show=False,
                                     **kargs):
        """
        函数返回：meteva形式的pandas.DataFrame格式；如果申明sava_path变量，可直接输出Micaps3文件
        ###
        qparams: 包含数据、列表、下载要素的字典容器
        time： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据
        userID: 业务账号，必须填
        pwd： 业务账户密码，必须填
        save_path: 存储文件目录，若不填写则不输出
        filename:  存储文件名,可以使用通配符(如YYYYMMDDHHFF.TTT即表示“年(四位)月(两位)日(两位)时(两位)分(两位).时效(三位)”；缺省-YYYYMMDDHHFF.000.m3)
        show_url:  是否完整显示url，默认为不显示
        defaultNan: 缺测显示值，默认缺测存成9999
        dataFormat: 返回数据格式，默认为json
        url: cmadaas的url前面固定部分，默认'http://10.40.17.54/music-ws/api?'
        SN_id: cmadaas服务节点代码，默认'NMIC_MUSIC_CMADAAS'
        ###
        """
        import os
        ## 根据参数字典，生成url
        LogLib.info('CMADaasAccess OBS output Starts: {0}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M")))
        time = CMADaasAccess.get_datetime_from_str(time)
        # print(time)
        urlstr = CMADaasAccess.combine_url_from_para(qparams, time, **kargs)
        if show: print(urlstr)
        ## 得到url返回结果
        json = CMADaasAccess.read_data_from_cmadaas(urlstr)
        ## 生成meteva数据
        sta = CMADaasAccess.dataframe_from_dict_json(json, default=defaultNan)
        ## 输出micaps3格式
        if save_path is not None:
            try:
                if filename is None: filename = "YYYYMMDDHHFF.000.m3"
                output_name = os.path.join(save_path, filename)
                output_name = meteva_base.get_path(output_name, time)
                print('OUTPUT: ', output_name)
                meteva_base.write_stadata_to_micaps3(sta, save_path=output_name, effectiveNum=2)
            except Exception as data:
                LogLib.error(
                    '{0} CMADaasAccess OBS output ERROR!: {1}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M"),
                                                                      data))
        return (sta)

    @staticmethod
    def get_files_from_cmadaas(qparams, time='202011110200', save_path=None, filename=None, show_files=True,
                               is_new=True,
                               **kargs):
        """
        time： 可以为字符串str，如“202011110200”；也可以为datetime类时间数据
        save_path: 存储文件目录，若不填写则不存储
        filename:  存储文件名,可以使用通配符(如YYYYMMDDHHFF.TTT即表示“年(四位)月(两位)日(两位)时(两位)分(两位).时效(三位)”；缺省-使用云平台原始文件名)
        show_url: 是否完整显示url，默认为不显示
        show_files: 是否完整显示待下载文件，默认为不显示
        is_new    ；True-只处理下载最新入库时间的一个文件;Fales-处理所有文件
        """
        import os
        import pandas as pd
        import datetime
        ## 根据参数字典，生成url
        time = CMADaasAccess.get_datetime_from_str(time)
        urlstr = CMADaasAccess.combine_url_from_para(qparams, time, **kargs)
        ## 得到url返回结果
        json = CMADaasAccess.read_data_from_cmadaas(urlstr)
        df = pd.DataFrame(json)
        try:
            df.sort_values(by='IYMDHM', inplace=True, ascending=False)  # 按入库时间降序排列，新更新的在上方
        except:
            print("no key IYMDHM")
        df.reset_index(drop=True, inplace=True)
        if is_new: df = df.loc[0].to_frame().T
        if show_files: print(df, type(df))

        try:
            LogLib.info('FixWriteData write_data_to_file start %s' % (save_path))
            if save_path is None: return (json)  # save_path为None，不存储文件，直接返回json信息

            for index in range(len(df)):
                fname = df.loc[index, 'FILE_NAME']
                url = df.loc[index, 'FILE_URL']
                datetime = df.loc[index, 'Datetime']
                ## url内容读取并文件保存
                file_time = CMADaasAccess.get_datetime_from_str(datetime)
                # 原url文件名
                if filename is None:
                    output_name = os.path.join(save_path, fname)
                # 自定义文件名格式
                else:
                    output_name = os.path.join(save_path, filename)
                output_name = meteva_base.get_path(output_name, file_time)
                # 文件存储
                grd, header = CMADaasAccess.read_file_from_cmadaas(url)
                print('WriteFile : ', output_name)
                CMADaasAccess.write_data_to_file(output_name, grd, creat_dir=True)

        except Exception as data:
            print("ERROR:", data)
            LogLib.error(
                '{0} CMADaasAccess Files output ERROR!: {1}'.format(datetime.datetime.now().strftime("%Y%m%d%H%M"),
                                                                    data))

        return (json)


#


if __name__ == '__main__':
    dati_list = [datetime.datetime(2021, 12, 1, 0, 0), datetime.datetime(2021, 12, 1, 1, 0)]
    qparams = {'serviceNodeId': 'NMIC_MUSIC_CMADAAS',
               'userId': 'NMC_SGFV',
               'interfaceId': 'getSateFileByTimeRange',
               'dataCode': 'SATE_FY4A_STA_L1',
               'timeRange': CMADaasAccess.get_timerange(dati_list),
               'pwd': 'NMC_yfsjyk_601'
               }
    save_dir = r"H:\test_data\output\meb\get_files"
    urlstr = CMADaasAccess.get_files_from_cmadaas(qparams, show_url=True, save_path=save_dir)
    print(urlstr)




# _*_ coding: utf-8 _*_

from os import path
from setuptools import find_packages, setup
from codecs import open

#name = 'meteva_base'
#author ="liucouhua,daikan,wangbaoli,tangbuxing"
#version =__import__(name).__version__
here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(  
    name = 'meteva_base',  
    version = '0.2.6',
    description = ("base classes and functions of generally meteorogical usage, like Basic_class/IO/calculation/plot .etc"),  
    long_description=long_description,
    long_description_content_type = 'text/markdown',
    # author
    author = 'NMC_WFT',  
    author_email = 'pangzide003@163.com',
    # license
    license = 'MIT', 
    classifiers=[
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Programming Language :: Python :: 3',
    ],
    #packages = ['meteva_base'],
    packages=find_packages(exclude=['docs', 'tests', 'build', 'dist']),
    include_package_data=True,
    exclude_package_data={'': ['.gitignore']},

    url = 'https://github.com/guoyunqian/meteva_base',
    install_requires=[
                      'numpy>=1.12.1,<=2.0.1',
                      'pandas>=1.0.4,<=2.2.2',
                      'netCDF4>=1.4.2,<=1.6.5',
                      'scipy>=1.0.0',
                      'xarray>=0.10.0, <=2024.7.0',
                      'scikit-learn>=1.0.0',
                      'matplotlib>=3.2.2',
                      'httplib2>=0.12.0',
                      'protobuf<=3.20.0',
                      'tables>=3.4.4',
                      'urllib3>=2.0',
                      'pyshp>=2.2.0',
                      'shapely>=2.0.0',
                      'h5py>=3.1.0',
## optional module
### contour functions
#                      "pyproj>=3.0.0",
#                      "fiona>=1.9.0",
#                      "geopandas>=0.14.0",
#                      "pyogrio>=0.7.0",
## lambert functions
#                      "cartopy>=0.20.0",
## grib support functions
#                      "eccodes>=1.5.0",
#                      "eccodes-python>=0.9.9",
#                      "cfgrib>=0.9.9",
                      ]
)

# development mode (DOS command):
#     python setup.py develop
#     python setup.py develop --uninstall

# build mode：
#     python setup.py build --build-base=D:/test/python/build

# distribution mode:
#     python setup.py sdist             # create source tar.gz file in /dist
#     python setup.py bdist_wheel       # create wheel binary in /dist
